
import jlogUtils = require('./jlog'); // log工具
import jsig_utils = require('./jsig_utils'); // 信号工具
import path = require('path'); // 路径工具

import Jlog = require('./jlog');
var jlog = Jlog("JaAddon");

/**
 * 插件工具类
 */
export = class JaAddon {

    /**
     * 载入的扩展Addon
     */
    public lib: any = null;

    /**
     * libPath路径
     */
    public libPath: string | null = null;

    /**
     * 内部对象
     */
    protected innerObj = {
        cbFunArr: <Array<any>>[], // 回调函数

        jlog: <any>undefined, // 日志工具

        // lib库记录日志函数
        libLog: (logLeave: number, logStr: string) => {
            var iobj = this.innerObj;
            switch (logLeave) {
                case 0:
                    iobj.jlog.trace(logStr);
                    break;
                case 1:
                    iobj.jlog.debug(logStr);
                    break;
                case 2:
                    iobj.jlog.info(logStr);
                    break;
                case 3:
                    iobj.jlog.warn(logStr);
                    break;
                case 4:
                    iobj.jlog.error(logStr);
                    break;
                default:
                    iobj.jlog.fatal(logStr);
                    break;
            }
        },

        // 载入lib
        loadLib: (libPath: string, libName: string) => {
            var self = this;
            var curLib = null;
            var libFullPath = "";
            libPath = path.normalize(libPath);

            // 错误处理
            var errDo = (err: any, path: string) => {
                if (err.code && err.code === "MODULE_NOT_FOUND") {
                    // 找不到文件
                    // if (jlog.jlogEnable) jlog.trace("Not find Lib:" + path);
                }
                else {
                    // 其他错误抛出
                    err.addonPath = path;
                    throw err;
                }
            };

            try {
                libFullPath = libPath + '/build/Debug/' + libName;
                curLib = require(libFullPath);
            } catch (error) {
                errDo(error, libFullPath);
            }
            if (!curLib) {
                try {
                    libFullPath = libPath + '/build/Release/' + libName;
                    curLib = require(libFullPath);
                } catch (error) {
                    errDo(error, libFullPath);
                }
            }
            if (!curLib) {
                try {
                    libFullPath = libPath + '/' + libName + '_' + process.platform + "_" + process.arch;
                    curLib = require(libFullPath);
                } catch (error) {
                    errDo(error, libFullPath);
                }
            }
            if (!curLib) {
                try {
                    libFullPath = libPath + '/' + libName;
                    curLib = require(libFullPath);
                } catch (error) {
                    errDo(error, libFullPath);
                }
            }

            if (curLib) {
                self.libPath = libFullPath;
                if (jlog.jlogEnable) jlog.info("Load Lib Success:" + libFullPath);
            }
            else {
                self.libPath = null;
            }

            return curLib;
        },

        // lib应答信号函数
        ackLibSig: (sigId: any, sigStr: string) => {
            jsig_utils.sigAck(sigId, JSON.parse(sigStr));
        }
    };

    /**
     * 构造函数
     */
    constructor() {

    }


    /**
     * 初始化
     * @param cfg 配置
     * libPath lib路径(省略后为当前项目的'../jaddon'或'./jmodel/jaddon_lib')
     * libName lib名称(省略后为'jaddon_lib')
     * logLevelSw 日志等级输出开关
     * @returns 加载的模块路径
     */
    init(cfg?: { libPath?: string, libName?: string, logLevelSw?: string }) {
        var self = this;
        var iobj = this.innerObj;

        // 检查值
        if (typeof (cfg) !== "object") cfg = {};
        var libPath = cfg.libPath ? cfg.libPath : process.cwd() + '/../jaddon';
        var libName = cfg.libName ? cfg.libName : 'jaddon_lib';
        var logLevelSw = cfg.logLevelSw ? cfg.logLevelSw : '111111';

        // 载入lib
        self.lib = iobj.loadLib(libPath, libName);
        if (!self.lib) {
            libPath = process.cwd() + '/jmodel/jaddon_lib';
            self.lib = iobj.loadLib(libPath, libName);
        }
        if (!self.lib) {
            return null;
        }

        // 日志工具
        iobj.jlog = jlogUtils(libName);

        // 初始化日志模块
        if (self.lib.extLogInit) {
            self.lib.extLogInit(logLevelSw, iobj.libLog);
        }

        // node信号回调初始化
        if (self.lib.extSigInit) {
            self.lib.extSigInit(iobj.ackLibSig);
        }

        // 初始化模块
        if (self.lib.addonsInit) {
            self.lib.addonsInit();
        }

        // 返回成功
        return self.libPath;
    }

    /**
     * 添加回调函数
     * @param cbFun 函数
     * @param funIndex 回调索引(0~N)
     * @param funName 函数名,为空返回函数定义名
     * @returns 回调函数个数
     */
    addCallBack(cbFun: string | Function, funIndex: number, funName?: string) {
        var self = this;
        var iobj = this.innerObj;
        if (!self.lib) return;

        // 函数判断
        if (typeof (cbFun) !== "function") return iobj.cbFunArr.length;

        // 函数索引
        if (!funIndex && funIndex !== 0) {
            funIndex = iobj.cbFunArr.length;
        }

        // 函数名
        if (!funName) funName = cbFun.name;

        // 添加函数
        iobj.cbFunArr[funIndex] = cbFun;

        // 放入lib
        self.lib.extAddCbFun(funIndex, cbFun, funName);
    }

    /**
     * 调试暂停
     * @returns 无
     */
    debugPause() {
        if (!this.lib) return;
        this.lib.extDebugPause();
    }

}
