"use strict";
/*

// 文档
https://knexjs.org/#Builder-transacting

// 普通查询
var a = yield mydb.q("select id from user where host=:name", { name: "张三" });

// 事务查询
yield mydb.t(function* (t,q) {
    var b = yield q("select 1");
    var c = yield q("select id from user where host=:name", { name: "张三" });
});

// 插入
yield mydb.q({
    sql: "insert into demo_table_xjz",
    insertVal: [
        {
            jid: base_parts.getId(),
            jtype: 1111,
            jname: 'nnnnn',
            jtime: new Date(),
            opUser:"ohehe",
        }
    ]
});

// 修改
yield mydb.q({
    sql: "update demo_table_xjz where jid>:jid",
    updateVal: {
        jtype: 222,
        jname: '🌮🌽🧑🏽🗡𓀫🤬👩🏼👩🏿🥟🍻🥠🔪🌯'
    }
},{
    jid:1,
    opUser:"ohehe",
});


// 分页查询
var out = yield db.q({
    sql: "select * from demo_table_xjz where jtime >:dateTime",
    pageOpt: {
        pageSize: 5,
        current: 1,
        needTotal: true
    },
    order: "jid desc",
    andWhere:[
        {
            field:"jid",
            op:">",
            value:1
        },
        {
            field:"jid",
            op:"is not null",
            value:""
        }
    ],

},{
    dateTime:'2018-11-21 17:49:22'
});

*/
const jbase_utils = require("./jbase_utils");
const jtype_utils = require("./jtype_utils");
const jret_utils = require("./jret_utils");
const jco = require("./jco");
const jmod_loader = require("./jmod_loader");
const Jlog = require("./jlog");
const JERR = require("./jerror_type");
var jlog = Jlog("JaDbKnex");
var Knex = require('knex');
/**
 * 获取表字符串
 * @param sqlStr SQL字段
 */
function getLastTabStr(sqlStr) {
    for (var index = sqlStr.length - 1; index >= 0; index--) {
        const ch = sqlStr.charCodeAt(index);
        if (ch === 45 || ch == 95 || (ch >= 48 && ch <= 57) || (ch >= 65 && ch <= 90) || (ch >= 97 && ch <= 122)) {
            continue;
        }
        else {
            return sqlStr.substr(index + 1);
        }
    }
    return sqlStr;
}
/**
 * 隐藏错误后抛出
 * @param err 原始错误
 */
function toHideErr(err) {
    if (jlog.jlogEnable)
        jlog.error(err);
    throw new Error("Database Query Error!");
}
module.exports = class JaDbKnex {
    // 构造
    constructor() {
        /**
         * 自动载入的模型
         */
        this.models = {};
        /**
         * 数据库对象
         */
        this.db = {};
        // 默认配置
        this.config = {
            "dialect": "postgres",
            "user": "postgres",
            "password": "",
            "host": "127.0.0.1",
            "database": "postgres",
            "pool": {
                "max": 50,
                "min": 1,
                "acquire": 40000,
                "idle": 35000
            },
            "storage": "./database.sqlite",
            "timezone": jbase_utils.getTimeZoneStr(),
            "errInfo": true,
            "logging": 0,
            "opField": false,
            "oracleConvert": true,
            "oracleTimeFormat": 'YYYY-MM-DD HH24:MI:SS.FF6',
            "timeSpTable": 0,
            "timeSpField": "jtime"
        };
        /**
         * 内部对象
         */
        this.innerObj = {
            /**
             * 时间分表清理对象
             */
            tSpTables: {},
            /**
             * 连接检查
             */
            chkLinkOk: null,
            /**
             * 是否需要Oracle转换
             */
            needOracleConvert: false,
            /**
             * 默认操作者
             */
            defaultOpUser: "quitter",
            /**
             * Oracle最后的Stream
             */
            lastOracleStream: null,
            /**
             * 数据库版本Int
             */
            sqlVerInt: 0,
            /**
             * 建立连接后执行函数
             */
            afterCreate: (conn, done) => {
                var self = this;
                var iobj = this.innerObj;
                // 连接成功
                var myDone = function (myErr, myConn) {
                    if (iobj.chkLinkOk) {
                        var chkLinkOk_t = iobj.chkLinkOk;
                        iobj.chkLinkOk = null;
                        chkLinkOk_t();
                    }
                    done(myErr, myConn);
                };
                // SQLServer设置
                if (self.config.dialect === 'mssql') {
                    if (iobj.sqlVerInt) {
                        myDone(null, conn);
                    }
                    else {
                        // 查询数据库版本
                        if (!conn.execSql) {
                            // 旧版本
                            conn.query(`select SERVERPROPERTY('ProductVersion') as productversion`, function (err, res) {
                                try {
                                    if (!iobj.sqlVerInt) {
                                        iobj.sqlVerInt = parseInt(res.recordset[0].productversion, 10);
                                        if (jlog.jlogEnable) {
                                            jlog.info("sqlVerInt:" + iobj.sqlVerInt);
                                        }
                                    }
                                }
                                catch (error) {
                                }
                                myDone(err, conn);
                            });
                        }
                        else {
                            // 新版本
                            var Request = require('tedious').Request;
                            var sqlRequest = new Request(`select SERVERPROPERTY('ProductVersion') as productversion`, function (err, rowCount) {
                                myDone(err, conn);
                            });
                            sqlRequest.on('row', function (columns) {
                                if (!iobj.sqlVerInt) {
                                    iobj.sqlVerInt = parseInt(columns[0].value, 10);
                                    if (jlog.jlogEnable) {
                                        jlog.info("sqlVerInt:" + iobj.sqlVerInt);
                                    }
                                }
                            });
                            conn.execSql(sqlRequest);
                        }
                    }
                }
                // Oracle设置
                else if (self.config.dialect === 'oracledb') {
                    // 接管Stream为了字段转换Oracle流补丁
                    var qStFun = conn.queryStream;
                    conn.queryStream = function myStFun() {
                        // 透传执行
                        var curSt = qStFun.apply(this, arguments);
                        // 接收元数据
                        curSt.once('metadata', function (metadata) {
                            curSt.metadataTmp = metadata;
                        });
                        // data转换
                        curSt.on('data', (data) => {
                            var outData = {};
                            if (curSt.metadataTmp) {
                                for (var i = 0; i < curSt.metadataTmp.length; i++) {
                                    const element = curSt.metadataTmp[i];
                                    // 去掉特殊字段
                                    if (element.name !== "ROWNUM_")
                                        outData[element.name] = data[i];
                                }
                                data.outData = outData;
                            }
                        });
                        // 返回
                        return curSt;
                    };
                    // 修改日期格式
                    conn.execute("ALTER SESSION SET NLS_TIMESTAMP_FORMAT = '" + self.config.oracleTimeFormat + "'", [], function (err, res) {
                        if (self.config.timezone) {
                            // 修改时区
                            conn.execute("ALTER SESSION SET TIME_ZONE = '" + self.config.timezone + "'", [], function (err, res) {
                                myDone(err, conn);
                            });
                        }
                        else {
                            myDone(err, conn);
                        }
                    });
                }
                // postgres设置
                else if (self.config.dialect === 'postgres') {
                    // 设置时区
                    if (self.config.timezone) {
                        var newTimezone = ((self.config.timezone[0] === '-') ? '+' : '-') + self.config.timezone.substr(1);
                        conn.query("SET timezone='" + newTimezone + "';", function (err) {
                            myDone(err, conn);
                        });
                    }
                    else {
                        myDone(null, conn);
                    }
                }
                // mysql设置
                else if (self.config.dialect === 'mysql' || self.config.dialect === 'mysql2') {
                    // 设置时区
                    if (self.config.timezone) {
                        conn.query("set time_zone = '" + self.config.timezone + "';", function (err) {
                            myDone(err, conn);
                        });
                    }
                    else {
                        myDone(null, conn);
                    }
                }
                else {
                    // 完成
                    myDone(null, conn);
                }
            },
            /**
             * 转换为Oracle名称
             * @param tbName 原始名称
             * @return 转换后的名称
             */
            toStdName: (tbName) => {
                var self = this;
                var iobj = this.innerObj;
                if (iobj.needOracleConvert) {
                    return tbName.replace(/\"/g, "").toUpperCase().trim();
                }
                return tbName.replace(/\"/g, "").trim();
            },
            /**
             * 转换为大写字段
             * @param Obj 原始对象
             * @return 转换后的对象
             */
            toObjBigField: (Obj, op_time, op_user) => {
                var retObj = {};
                for (const key in Obj) {
                    retObj[key.toUpperCase()] = Obj[key];
                }
                if (op_time) {
                    retObj["OP_TIME"] = op_time;
                    retObj["OP_USER"] = op_user;
                }
                return retObj;
            },
            /**
             * 转换Update对象
             * @param updateVal Insert对象
             * @param op_user 操作者
             */
            toUpdateVal: (updateVal, op_user) => {
                var self = this;
                var iobj = this.innerObj;
                // 操作字段
                var op_time = null;
                if (self.config.opField) {
                    op_time = new Date();
                }
                if (iobj.needOracleConvert) {
                    return iobj.toObjBigField(updateVal, op_time, op_user);
                }
                else if (op_time) {
                    updateVal["op_time"] = op_time;
                    updateVal["op_user"] = op_user;
                }
                return updateVal;
            },
            /**
             * 转换Insert对象
             * @param insertVal Insert对象
             * @param op_user 操作者
             */
            toInsertVal: (insertVal, op_user) => {
                var self = this;
                var iobj = this.innerObj;
                // 操作字段
                var op_time = null;
                if (self.config.opField) {
                    op_time = new Date();
                }
                if (iobj.needOracleConvert) {
                    // Oracle特殊处理
                    var ret = [];
                    for (var index = 0; index < insertVal.length; index++) {
                        var newObj = iobj.toObjBigField(insertVal[index], op_time, op_user);
                        ret.push(newObj);
                    }
                    return ret;
                }
                else {
                    // 其他数据库
                    if (self.config.opField) {
                        for (var index = 0; index < insertVal.length; index++) {
                            var curField = insertVal[index];
                            if (op_time) {
                                curField["op_time"] = op_time;
                                curField["op_user"] = op_user;
                            }
                        }
                    }
                    return insertVal;
                }
            },
            /**
              * 记录转换
              */
            rowConvert: (rowData, param, fieldFmt, isTest) => {
                var self = this;
                var iobj = this.innerObj;
                // 初始化
                if (!fieldFmt.initOk) {
                    fieldFmt.initOk = 1;
                    // 空白记录
                    if (!rowData)
                        return;
                    // 时间转换格式选项
                    var fieldFormat = (param.fieldFormat === undefined) ? self.config.fieldFormat : param.fieldFormat;
                    // 处理字段
                    for (const key in rowData) {
                        // Oracle字段转换
                        var keyLow = key.toLowerCase();
                        if (iobj.needOracleConvert && (key != keyLow)) {
                            fieldFmt.initOk = 2; // 需要转换
                            if (!fieldFmt.fieldN2N)
                                fieldFmt.fieldN2N = {};
                            fieldFmt.fieldN2N[key] = keyLow;
                        }
                        // 字段格式转换
                        if (fieldFormat) {
                            if (!fieldFmt.fieldN2N)
                                fieldFmt.fieldN2N = {};
                            if (fieldFormat === 1) {
                                fieldFmt.fieldN2N[key] = jbase_utils.u2h(keyLow);
                            }
                            else if (fieldFormat === 2) {
                                fieldFmt.fieldN2N[key] = jbase_utils.h2u(key);
                            }
                            else if (fieldFormat === 3) {
                                fieldFmt.fieldN2N[key] = keyLow;
                            }
                            else if (fieldFormat === 4) {
                                fieldFmt.fieldN2N[key] = key.toUpperCase();
                            }
                            // 删除无用的转换
                            if (fieldFmt.fieldN2N[key] === key) {
                                delete fieldFmt.fieldN2N[key];
                            }
                            else {
                                fieldFmt.initOk = 2; // 需要转换
                            }
                        }
                        // 时间格式转换
                        if (param.timeFmt && (Object.prototype.toString.call(rowData[key]) === '[object Date]')) {
                            fieldFmt.initOk = 2; // 需要转换
                            if (!fieldFmt.fieldT2S)
                                fieldFmt.fieldT2S = {};
                            fieldFmt.fieldT2S[key] = param.timeFmt;
                        }
                    }
                    if (isTest)
                        return;
                }
                // 转换时间
                if (fieldFmt.fieldT2S) {
                    for (const key in fieldFmt.fieldT2S) {
                        if (fieldFmt.fieldT2S[key] === 2) {
                            rowData[key] = jtype_utils.toDateStr(rowData[key]);
                        }
                        else {
                            rowData[key] = jtype_utils.getDate(rowData[key], undefined, fieldFmt.fieldT2S[key]);
                        }
                    }
                }
                // 转换字段
                if (fieldFmt.fieldN2N) {
                    for (const key in fieldFmt.fieldN2N) {
                        rowData[fieldFmt.fieldN2N[key]] = rowData[key];
                        delete rowData[key];
                    }
                }
            },
            /**
             * 原始查询
             * @param kObj 查询对象
             * @param opType 查询类型
             * @param sql SQL语句
             * @paramOpt paramOpt 绑定及参数(分页参数在里面是无效的)
             * @param pageOpt 分页参数
             */
            queryEx: (kObj, opType, sql, paramOpt, paramData, rowfun, pageOpt, selectStream) => {
                var self = this;
                var iobj = this.innerObj;
                var tableName = "";
                // 超时
                var curTimeout = paramOpt.queryTimeout || self.config.jtimeOut;
                // 插入
                if (opType === 'insert' && paramOpt.insertVal) {
                    // 强制为数组
                    var insertVal = jtype_utils.getArray(paramOpt.insertVal, [paramOpt.insertVal]);
                    // 取出表名
                    var pos_t = sql.toLowerCase().indexOf("into ");
                    if (pos_t >= 0 && pos_t < 2) {
                        tableName = sql.substr(pos_t + "into ".length);
                    }
                    else {
                        tableName = sql;
                    }
                    tableName = iobj.toStdName(tableName);
                    // 查询体
                    var retInsertQuery = kObj(tableName).insert(iobj.toInsertVal(insertVal, paramOpt.opUser || iobj.defaultOpUser));
                    if (self.config.logging === 2) {
                        if (jlog.jlogEnable)
                            jlog.info("querySQL:" + retInsertQuery.toQuery());
                    }
                    // 返回
                    return retInsertQuery;
                }
                // 修改
                else if (opType === 'update' && paramOpt.updateVal) {
                    // 取出表名
                    var pos_t = (sql + " ").toLowerCase().indexOf(" ");
                    var whereTxt;
                    if (pos_t >= 0) {
                        tableName = sql.substr(0, pos_t);
                        sql = sql.substr(pos_t).trim();
                    }
                    else {
                        throw new Error("unknow tableName");
                    }
                    tableName = iobj.toStdName(tableName);
                    // 取出where条件
                    pos_t = sql.toLowerCase().indexOf("where");
                    if (pos_t >= 0 && pos_t < 2) {
                        whereTxt = sql.substr(pos_t + "where".length).trim();
                    }
                    // 修改
                    var retUpdateQuery = kObj(tableName).update(iobj.toUpdateVal(paramOpt.updateVal, paramOpt.opUser || iobj.defaultOpUser));
                    if (whereTxt)
                        retUpdateQuery = retUpdateQuery.whereRaw(whereTxt, paramData);
                    if (self.config.logging === 2) {
                        if (jlog.jlogEnable)
                            jlog.info("querySQL:" + retUpdateQuery.toQuery());
                    }
                    return retUpdateQuery;
                }
                // 查询
                else if (opType === "select") {
                    var retSelectQuery;
                    var isYieldCbFun = jbase_utils.isYieldFun(rowfun);
                    if (pageOpt) {
                        // 判断页码
                        if (pageOpt.current <= 0) {
                            throw new Error("pageOpt.current must 1~N");
                        }
                        // 判断是否存在order
                        var indexOrder = sql.toLowerCase().lastIndexOf(" order ");
                        if (indexOrder <= 0) {
                            throw new Error('no order !');
                        }
                        // 计算分页开始行(0~N)
                        var curNum = (pageOpt.current - 1) * pageOpt.pageSize;
                        // 判断SQLserver和版本
                        if ((curNum > 0) && (self.config.dialect === 'mssql') && (iobj.sqlVerInt < 11)) {
                            // 当不支持分页语法时
                            retSelectQuery = kObj.select(kObj.raw(`* from (select ROW_NUMBER() OVER(${sql.substr(indexOrder)}) AS jpage_row_id,${sql.substr(0, indexOrder)}) as jpage_t where jpage_t.jpage_row_id between ${curNum + 1} and ${curNum + pageOpt.pageSize} `, paramData));
                        }
                        else {
                            // 分页查询
                            retSelectQuery = kObj.select(kObj.raw(sql, paramData)).limit(pageOpt.pageSize).offset(curNum);
                        }
                    }
                    else {
                        // 普通查询
                        retSelectQuery = kObj.select(kObj.raw(sql, paramData));
                    }
                    // 设置超时
                    if (curTimeout) {
                        retSelectQuery.timeout(curTimeout);
                    }
                    // 打印日志
                    if (self.config.logging === 2) {
                        if (jlog.jlogEnable)
                            jlog.info("querySQL:" + retSelectQuery.toQuery());
                    }
                    // 流读取
                    if (selectStream) {
                        if (rowfun) {
                            if (typeof (selectStream) !== 'object') {
                                selectStream = {};
                            }
                            if (!selectStream.highWaterMark)
                                selectStream.highWaterMark = 5000;
                            var streamObj = retSelectQuery.stream(selectStream);
                            return new Promise((resolve, reject) => {
                                var curIndex = 0;
                                var busy = 0;
                                var fieldFmt = {};
                                streamObj.on('data', (data) => {
                                    // Oracle流Bug兼容
                                    var rowData = data;
                                    if (rowData.outData) {
                                        rowData = rowData.outData;
                                        delete rowData.outData;
                                    }
                                    // 字段转换
                                    if (fieldFmt.initOk !== 1) {
                                        iobj.rowConvert(rowData, paramOpt, fieldFmt);
                                    }
                                    // 存在回调函数
                                    if (isYieldCbFun) {
                                        busy = 1;
                                        streamObj.pause();
                                        jco(rowfun(rowData, curIndex)).then((res) => {
                                            if (busy === 2) {
                                                resolve();
                                            }
                                            busy = 0;
                                            streamObj.resume();
                                        }).catch((err) => {
                                            if (jlog.jlogEnable)
                                                jlog.error("user code err:", curIndex, rowData, err);
                                            if (busy === 2) {
                                                resolve();
                                            }
                                            busy = 0;
                                            streamObj.resume();
                                        });
                                    }
                                    else {
                                        try {
                                            rowfun(rowData, curIndex);
                                        }
                                        catch (error) {
                                            if (jlog.jlogEnable)
                                                jlog.error("user code err:", curIndex, rowData, error);
                                        }
                                    }
                                    curIndex++;
                                });
                                streamObj.once('end', () => {
                                    if (busy === 1) {
                                        // 异步未完成,交给异步处理
                                        busy = 2;
                                    }
                                    else {
                                        resolve();
                                    }
                                });
                                streamObj.on('error', (err) => {
                                    reject(err);
                                });
                            });
                        }
                        else {
                            throw new Error('select Stream must use cbFun!');
                        }
                    }
                    // 处理结果
                    return jco(function* () {
                        // 获取结果集数组
                        var getObj = yield retSelectQuery;
                        // 转换测试
                        var fieldFmt = {};
                        iobj.rowConvert(getObj[0], paramOpt, fieldFmt, true);
                        // 其他查询字段转换
                        if (fieldFmt.initOk === 2 && rowfun) {
                            for (var i = 0; i < getObj.length; i++) {
                                // 转换类型
                                iobj.rowConvert(getObj[i], paramOpt, fieldFmt);
                                // 回调
                                try {
                                    if (isYieldCbFun) {
                                        yield rowfun(getObj[i], i, getObj);
                                    }
                                    else {
                                        rowfun(getObj[i], i, getObj);
                                    }
                                }
                                catch (error) {
                                    if (jlog.jlogEnable)
                                        jlog.error("user code err:", i, getObj[i], error);
                                }
                            }
                        }
                        // 仅转换
                        else if (fieldFmt.initOk === 2) {
                            for (var i = 0; i < getObj.length; i++) {
                                // 转换类型
                                iobj.rowConvert(getObj[i], paramOpt, fieldFmt);
                            }
                        }
                        // 仅回调
                        else if (rowfun) {
                            for (var i = 0; i < getObj.length; i++) {
                                try {
                                    if (isYieldCbFun) {
                                        yield rowfun(getObj[i], i, getObj);
                                    }
                                    else {
                                        rowfun(getObj[i], i, getObj);
                                    }
                                }
                                catch (error) {
                                    if (jlog.jlogEnable)
                                        jlog.error("user code err:", i, getObj[i], error);
                                }
                            }
                        }
                        // 不需要转换
                        return getObj;
                    });
                }
                else {
                    // 警告原始修改
                    if (self.config.opField) {
                        if (opType === 'insert') {
                            if (jlog.jlogEnable)
                                jlog.warn('opField is enable,please use [insertVal] or db.insert() to insert');
                        }
                        else if (opType === 'update') {
                            if (jlog.jlogEnable)
                                jlog.warn('opField is enable,please use [updateVal] or db.update() to update');
                        }
                    }
                    // 查询体
                    var retQuery = kObj.raw(opType + " " + sql, paramData);
                    // 设置超时
                    if (curTimeout) {
                        retQuery.timeout(curTimeout, {});
                    }
                    if (self.config.logging === 2) {
                        if (jlog.jlogEnable)
                            jlog.info("querySQL:" + retQuery.toQuery());
                    }
                    // 普通查询
                    return retQuery;
                }
            }
        };
    }
    /**
     * 初始化
     * @param userCfg 配置
     * @returns 未定义
     */
    init(userCfg) {
        var self = this;
        var iobj = this.innerObj;
        return jco(function* () {
            // knex独有配置
            var knexConfig = {
                connection: {},
                pool: {
                    afterCreate: iobj.afterCreate // 创建后任务
                },
                // 连接池取出连接超时ms
                acquireConnectionTimeout: 120000,
                // // 结果后置处理
                // postProcessResponse: (result, queryContext) => {
                //     return result;
                // },
            };
            // 合并配置
            jbase_utils.unionObj(self.config, userCfg);
            // 配置容错
            if (self.config.dialect === "mysql")
                self.config.dialect = "mysql2"; // 目前mysql库支持不好,强制使用mysql2
            // 配置转换
            knexConfig.client = self.config.dialect;
            knexConfig.debug = (self.config.logging === 1 || self.config.logging === true) ? true : false;
            var connection = knexConfig.connection;
            if (self.config.host)
                connection.host = self.config.host;
            if (self.config.port)
                connection.port = self.config.port;
            if (self.config.user)
                connection.user = self.config.user;
            if (self.config.password)
                connection.password = self.config.password;
            if (self.config.database)
                connection.database = self.config.database;
            if (self.config.options)
                connection.options = self.config.options;
            if (knexConfig.pool && self.config.pool) {
                if (self.config.pool.min)
                    knexConfig.pool.min = self.config.pool.min;
                if (self.config.pool.max)
                    knexConfig.pool.max = self.config.pool.max;
                if (self.config.pool.acquire)
                    knexConfig.pool.acquireTimeoutMillis = self.config.pool.acquire;
                if (self.config.pool.idle)
                    knexConfig.pool.idleTimeoutMillis = self.config.pool.idle;
            }
            // op操作参数
            if (self.config.opField && self.config.opField.defaultOpUser)
                iobj.defaultOpUser = self.config.opField.defaultOpUser;
            // mssql差异配置
            if (self.config.dialect === 'mssql') {
                // 超时设置
                if (connection.requestTimeout === undefined)
                    connection.requestTimeout = self.config.jtimeOut || (60 * 60 * 1000);
                // 选项
                if (connection.options === undefined)
                    connection.options = {};
                // 加密连接(云平台通常使用加密,但是本地数据框一般都是非加密连接)
                if (connection.options.encrypt === undefined)
                    connection.options.encrypt = false;
                // Date类型使用软件时区
                if (connection.options.useUTC === undefined)
                    connection.options.useUTC = false;
                // 主机包含端口,优先使用
                if (connection.host) {
                    if (connection.host.indexOf(',') > 0) {
                        var hostSp = connection.host.split(',');
                        connection.host = hostSp[0];
                        connection.port = parseInt(hostSp[1], 10);
                    }
                }
            }
            // oracle差异配置
            if (self.config.dialect === "oracledb") {
                iobj.needOracleConvert = (self.config.oracleConvert) ? true : false;
                // 配置
                if (self.config.initOracleCfg) {
                    try {
                        var oracledb = require('oracledb');
                        oracledb.initOracleClient(self.config.initOracleCfg);
                    }
                    catch (err) {
                        if (jlog.jlogEnable)
                            jlog.error(err);
                    }
                }
            }
            // 日志
            knexConfig.log = {
                warn(message) {
                    if (jlog.jlogEnable)
                        jlog.warn(self.config.host, message);
                },
                error(message) {
                    if (jlog.jlogEnable)
                        jlog.error(self.config.host, message);
                },
                deprecate(message) {
                    if (jlog.jlogEnable)
                        jlog.warn(self.config.host, message);
                },
                debug(message) {
                    if (jlog.jlogEnable)
                        jlog.debug(self.config.host, message);
                },
            };
            // 合并扩展配置
            jbase_utils.unionObj(knexConfig, userCfg.configEx);
            // 创建数据库类
            if (jlog.jlogEnable)
                jlog.info("Load Database:dialect=" + self.config.dialect + ",database=" + self.config.database + ",user=" + self.config.user);
            var knex = Knex(knexConfig);
            // 测试数据库连接
            while (true) {
                try {
                    yield new Promise((resolve, reject) => {
                        // 已决函数
                        iobj.chkLinkOk = resolve;
                        // 空查询
                        knex.raw("    ")
                            .then((res) => {
                            // 成功,这里不做处理
                        })
                            .catch((err) => {
                            // 错误且没用成功
                            if (iobj.chkLinkOk) {
                                iobj.chkLinkOk = null;
                                reject(err);
                            }
                        });
                    });
                    if (jlog.jlogEnable)
                        jlog.info("Test Database[" + self.config.database + "] Success");
                    break;
                }
                catch (error) {
                    if (jlog.jlogEnable)
                        jlog.error("Test Database[" + self.config.database + "] Error:", error);
                    if (!self.config.firstTryCount)
                        self.config.firstTryCount = 0;
                    self.config.firstTryCount--;
                    if (self.config.firstTryCount === -1) {
                        throw error;
                    }
                    yield jbase_utils.sleep(5000);
                }
            }
            // 成功,功能重定向
            self.db = knex;
        });
    }
    /**
     * 数据库事务
     * @param tranFun 事务函数
     * @returns 无
     */
    t(tranFun) {
        var self = this;
        var tranRet = self.db.transaction((trx) => {
            return jco(tranFun(trx, function (a, b, c) {
                // 转换为事物查询,末尾带入事物函数
                return self.q(a, b, c, trx);
            }));
        });
        return tranRet;
    }
    /**
     * 查询数据库
     * @param sql_opt SQL语句或param选项
     * @param sqlParam 数据和选项(当sql_opt为对象时sqlParam仅做数据使用)
     * @param rowfun 处理函数fuction(row,index,all_rows)
     * @returns 查询对象
     */
    select(sql_opt, sqlParam, rowfun) {
        var self = this;
        return self.q(sql_opt, sqlParam, rowfun);
    }
    /**
     * 插入
     * @param tableName 表名
     * @param insertVal 插入数据,可以是对象,或者是数组(多条)
     * @param opt 选项
     * @param opUser 操作用户
     */
    insert(tableName, insertVal, opt, opUser) {
        var self = this;
        return self.q({
            sql: "insert into " + tableName,
            insertVal: insertVal,
            opUser: opUser
        }, opt);
    }
    /**
     * 修改
     * @param tableName 表名
     * @param updateVal 修改的数据
     * @param where where条件,如:"abc=123"
     * @param param 注入参数
     * @param opt_opUser 选项或者操作者
     * @param opUser 操作用户
     */
    update(tableName, updateVal, where, param, opt_opUser, opUser) {
        var self = this;
        // 位置转换
        if (typeof (opt_opUser) === 'string') {
            opUser = opt_opUser;
            opt_opUser = {};
        }
        var myOpt = {
            sql: "update " + tableName + (where ? " where " + where : ""),
            updateVal: updateVal,
            opUser: opUser
        };
        jbase_utils.unionObj(myOpt, opt_opUser);
        return self.q(myOpt, param);
    }
    /**
     * 删除
     * @param tableName 表名
     * @param where where条件,如:"abc=123"
     * @param param 注入参数
     * @param opt 选项
     */
    delete(tableName, where, param, opt) {
        var self = this;
        var myOpt = {
            sql: "delete from " + tableName + (where ? " where " + where : ""),
        };
        jbase_utils.unionObj(myOpt, opt);
        return self.q(myOpt, param);
    }
    /**
     * 查询数据库
     * @param sql_opt SQL语句或param参数
     * @param sqlParam 参数,如果sql_opt也有参数,则两者将会合并
     * @param rowfun 处理函数fuction(row,index,all_rows)
     * @returns 查询对象
     */
    q(sql_opt, sqlParam, rowfun) {
        var self = this;
        var iobj = this.innerObj;
        // 获取快照
        var lastStack = {};
        if (self.config.errInfo) {
            Error.captureStackTrace(lastStack);
        }
        if (!self.db) {
            return Promise.reject(new Error("no init!"));
        }
        // 事物查询体
        var trx = arguments[3];
        // 异步查询
        return jco(function* () {
            // 参数处理
            var sql;
            var paramOpt = {};
            ;
            var paramData = {};
            // 参1字符串,参2选项与数据合并
            if (typeof (sql_opt) === "string") {
                sql = sql_opt;
                jbase_utils.unionObj(paramOpt, sqlParam, 0);
                paramData = paramOpt;
            }
            // 参1作为选项,参2仅作为数据
            else if (sqlParam) {
                sql = sql_opt.sql || "";
                jbase_utils.unionObj(paramOpt, sql_opt, 0);
                jbase_utils.unionObj(paramData, sqlParam, 0);
            }
            // 参1作为选项,无参2
            else {
                // 独立对象参数
                sql = sql_opt.sql || "";
                jbase_utils.unionObj(paramOpt, sql_opt, 0);
            }
            if (!sql)
                sql = paramOpt.sql || "";
            // 展开SQL
            var lineSql = sql.replace(/(\r|\n)/g, " ").replace(/\(/g, " ( ").replace(/\)/g, " ) ").trim();
            lastStack.lineSql = lineSql;
            // 操作方式
            var sql_op = lineSql.substr(0, 10).toLowerCase();
            var opType = "select";
            var opPos = sql_op.indexOf(opType);
            if (opPos < 0) {
                opType = "insert";
                opPos = sql_op.indexOf(opType);
                if (opPos < 0) {
                    opType = "delete";
                    opPos = sql_op.indexOf(opType);
                    if (opPos < 0) {
                        opType = "update";
                        opPos = sql_op.indexOf(opType);
                        if (opPos < 0) {
                            opType = "select";
                            opPos = sql_op.indexOf(opType);
                        }
                    }
                }
            }
            // 裁剪SQL
            if (opPos >= 0) {
                lineSql = lineSql.substr(opPos + opType.length).trim();
            }
            // 事务体
            if (paramOpt.transaction && paramOpt.transaction.commit) {
                // 优先主参数
                trx = paramOpt.transaction;
            }
            // 查询体
            var queryFun = trx || self.db;
            // 查询
            var ret = [];
            var total = -1;
            // 分页信息
            var pageOpt = (paramOpt.pageOpt) ? {
                pageSize: jtype_utils.getInt(paramOpt.pageOpt.pageSize, 30),
                current: jtype_utils.getInt(paramOpt.pageOpt.current, 1),
                needTotal: paramOpt.pageOpt.needTotal
            } : undefined;
            try {
                // andWhere
                if (paramOpt.andWhere && paramOpt.andWhere.length) {
                    var sqlAndWhere = "";
                    for (var index = 0; index < paramOpt.andWhere.length; index++) {
                        const element = paramOpt.andWhere[index];
                        var covStr = (typeof (element.value) === "string") ? "'" : "";
                        if (sqlAndWhere.length > 0)
                            sqlAndWhere += " and ";
                        switch (element.op) {
                            case ">":
                            case "<":
                            case ">=":
                            case "<=":
                            case "=":
                            case "!=":
                            case "like":
                                sqlAndWhere += "(" + jbase_utils.filterSQL(element.field) + " " + element.op + " " + covStr + jbase_utils.filterSQL(element.value + "") + covStr + ")";
                                break;
                            case "is null":
                            case "is not null":
                                sqlAndWhere += "(" + jbase_utils.filterSQL(element.field) + " " + element.op + " )";
                                break;
                            default:
                        }
                    }
                    // 合并语句
                    if (sqlAndWhere) {
                        var sqlLc = lineSql.toLowerCase();
                        var indexWhere = sqlLc.indexOf(" where ");
                        var indexOrder = sqlLc.indexOf(" order ");
                        // 拆开order
                        var orderFirst;
                        var orderLast;
                        if (indexOrder > 0) {
                            orderFirst = lineSql.substr(0, indexOrder);
                            orderLast = lineSql.substr(indexOrder);
                        }
                        else {
                            orderFirst = lineSql;
                            orderLast = "";
                        }
                        // 存在where字段
                        if (indexWhere > 0) {
                            lineSql = orderFirst + " and ( " + sqlAndWhere + " ) " + orderLast;
                        }
                        else {
                            lineSql = orderFirst + " where " + sqlAndWhere + " " + orderLast;
                        }
                    }
                }
                // 时间分表
                if (!paramOpt.noUseTimeSpTable && self.config.timeSpTable) {
                    // 分析是否存在
                    var sqlSp = lineSql.split('_xjz');
                    if (sqlSp.length > 1) {
                        // 当前时间
                        var curIntTime = Date.now();
                        var cutTime = curIntTime - (self.config.timeSpTable * 24 * 60 * 60 * 1000);
                        // 提取表名称
                        for (var index = 0; index < sqlSp.length - 1; index++) {
                            var tableName = getLastTabStr(sqlSp[index]);
                            if (!tableName)
                                continue;
                            // 没有则新建
                            if (!iobj.tSpTables[tableName]) {
                                iobj.tSpTables[tableName] = {
                                    lastClean: 0 // 立即开始
                                };
                            }
                            // 清理
                            if ((curIntTime - iobj.tSpTables[tableName].lastClean) > (24 * 60 * 60 * 1000)) {
                                iobj.tSpTables[tableName].lastClean = curIntTime;
                                jco(function* () {
                                    yield jbase_utils.sleep(30 * 1000);
                                    if (jlog.jlogEnable)
                                        jlog.info(`db clean table=${tableName}_tmp`);
                                    try {
                                        var delTime = new Date(curIntTime - ((self.config.timeSpTable + 1) * 24 * 60 * 60 * 1000));
                                        yield self.q(`delete from ${tableName}_tmp where ${self.config.timeSpField} < :mydate`, {
                                            noUseTimeSpTable: true,
                                            mydate: delTime
                                        });
                                        if (jlog.jlogEnable)
                                            jlog.info(`db clean table=${tableName}_tmp finish`);
                                    }
                                    catch (error) {
                                        if (jlog.jlogEnable)
                                            jlog.error(`db clean table=${tableName}_tmp`);
                                    }
                                });
                            }
                        }
                        // 查询类,根据时间字段选择表
                        if (opType === 'select') {
                            var dateTime = jtype_utils.getDate(paramOpt.dateTime, cutTime, 1);
                            var beginTime = jtype_utils.getDate(paramOpt.beginTime, cutTime, 1);
                            var endTime = jtype_utils.getDate(paramOpt.endTime, cutTime, 1);
                            // 时间段有效
                            if ((dateTime >= cutTime) || (beginTime >= cutTime)) {
                                // 使用快速表
                                lineSql = lineSql.replace(/_xjz/g, "_tmp");
                            }
                            else if ((dateTime === cutTime) && (beginTime === cutTime) && (endTime === cutTime)) {
                                // 无任何限制条件
                                throw Error("Time Split Table param no [dateTime/beginTime/endTime]");
                            }
                        }
                        // 改写,双表操作
                        else {
                            ret = yield iobj.queryEx(queryFun, opType, lineSql.replace(/_xjz/g, "_tmp"), paramOpt, paramData).catch(toHideErr);
                        }
                    }
                }
                // 分页总数查询
                if (pageOpt && opType === 'select') {
                    var fromIndex = lineSql.toLowerCase().indexOf(" from ");
                    if (fromIndex < 0) {
                        throw new Error("pageOpt err,nofind:from,sql=" + lineSql);
                    }
                    // 查询总数
                    if (pageOpt.needTotal && pageOpt.current === 1) {
                        var retCountList = yield iobj.queryEx(queryFun, "select", `count(1) as "count" ` + lineSql.substr(fromIndex), paramOpt, paramData).catch(toHideErr);
                        if (retCountList.length > 0) {
                            total = jtype_utils.getInt(parseInt(retCountList[0].count, 10), 0);
                        }
                    }
                }
                // 加入order
                if (paramOpt.order && typeof (paramOpt.order) === "string") {
                    lineSql = lineSql + " order by " + jbase_utils.filterSQL(paramOpt.order);
                }
                // 主查询
                ret = yield iobj.queryEx(queryFun, opType, lineSql, paramOpt, paramData, rowfun, pageOpt, paramOpt.selectStream).catch(toHideErr);
                // 只要第一行数据
                if (paramOpt.onlyOne) {
                    if (!ret.length)
                        return null;
                    return ret[0];
                }
                // 分页返回
                if (pageOpt) {
                    // 添加分页信息
                    var pageRet = {
                        list: ret
                    };
                    jret_utils.addPageInfo(pageRet, total, pageOpt.current, pageOpt.pageSize);
                    return pageRet;
                }
                else {
                    // 直接返回
                    return ret;
                }
            }
            catch (error) {
                // 额外的错误信息
                if (lastStack && lastStack.stack) {
                    error.lastStack = lastStack.stack;
                    error.lineSql = lastStack.lineSql;
                }
                // 错误输出
                if (jlog.jlogEnable)
                    jlog.error("error:", error);
                throw error;
            }
        });
    }
    /**
     * 通过定义创建表,若存在则不创建
     * @param defineObj 对象定义
     */
    createByDefine(defineObj) {
        var self = this;
        var iobj = this.innerObj;
        return jco(function* () {
            // 表名
            var tableName = iobj.toStdName(defineObj.tableName);
            // 检查是否存在表
            var hasTable = yield self.db.schema.hasTable(tableName);
            if (!hasTable) {
                // 创建表
                if (jlog.jlogEnable)
                    jlog.info("create table:" + tableName);
                yield self.db.schema.createTable(tableName, function (t) {
                    // 字段解析
                    for (const key in defineObj.attObj) {
                        const defField = defineObj.attObj[key];
                        var fieldName = iobj.toStdName(key);
                        // 类型转换
                        if (typeof (defField.type) === "function")
                            defField.type = defField.type();
                        if (typeof (defField.type) !== "string")
                            defField.type = "";
                        defField.type += ",";
                        // 设置类型
                        var tp = defField.type.split(",");
                        var type = tp[0];
                        var len = tp[1] ? parseInt(tp[1], 10) : 0;
                        var outField;
                        if (type === "BIGINT") {
                            outField = t.bigInteger(fieldName);
                        }
                        else if (type === "INTEGER") {
                            outField = t.integer(fieldName);
                        }
                        else if (type === "DATE") {
                            outField = t.dateTime(fieldName);
                        }
                        else if (type === "STRING") {
                            outField = t.string(fieldName, len);
                        }
                        else {
                            throw new Error("field type is Err,must:BIGINT|INTEGER|DATE|STRING(n)");
                        }
                        // 主键
                        if (defField.primaryKey)
                            outField = outField.primary();
                        // 是否为空
                        if (defField.allowNull) {
                            outField = outField.nullable();
                        }
                        else {
                            outField = outField.notNullable();
                        }
                        // 默认值
                        if (defField.defaultValue !== undefined) {
                            outField = outField.defaultTo(defField.defaultValue);
                        }
                    }
                    // 加入时间戳字段
                    if (self.config.opField) {
                        t.dateTime(iobj.toStdName("op_time")).nullable();
                        t.string(iobj.toStdName("op_user"), 100).nullable();
                    }
                    // 选项
                    var opt = defineObj.opt;
                    // 索引
                    if (opt.indexes) {
                        for (var ii = 0; ii < opt.indexes.length; ii++) {
                            const curIndex = opt.indexes[ii];
                            var indexList = [];
                            for (var iii = 0; iii < curIndex.fields.length; iii++) {
                                indexList.push(iobj.toStdName(curIndex.fields[iii]));
                            }
                            // 索引
                            t.index(indexList);
                            // 唯一键
                            if (curIndex.unique)
                                t.unique(indexList);
                        }
                    }
                });
                // 执行表数据
                var dataSql = jtype_utils.getArray(defineObj.opt.dataSql, null, 1);
                if (dataSql) {
                    try {
                        if (jlog.jlogEnable)
                            jlog.info("insert data to table:" + tableName + ",count=" + dataSql.length);
                        yield self.db.insert(iobj.toInsertVal(dataSql, "CreateTable")).into(tableName);
                    }
                    catch (err1) {
                        if (jlog.jlogEnable)
                            jlog.error("dataSql [" + tableName + "] Error:", err1);
                        throw err1;
                    }
                }
            }
        });
    }
    /**
     * 通过模型创造表定义
     * @param dbModObj 模型对象
     * @param optCbFun 选项回调函数
     */
    makeDefineByMod(dbModObj, optCbFun) {
        var self = this;
        // 创建对象
        var curObj1 = dbModObj({
            define: (tableName, attObj, opt) => {
                // 返回体
                var cbObj = {
                    tableName,
                    attObj,
                    opt: opt ? opt : {}
                };
                // 自定义回调处理
                if (optCbFun)
                    optCbFun(cbObj);
                return cbObj;
            }
        }, self);
        // 放入模型表
        self.models[curObj1.name] = curObj1;
        return curObj1;
    }
    /**
     * 载入模型,将会在models产生模型对象
     * @param modelDir 模型目录
     * @param createTable 自动创建表
     * @returns 模型原始定义对象
     */
    loadModel(modelDir, createTable) {
        var self = this;
        var iobj = this.innerObj;
        return jmod_loader.loadAllModules(modelDir, false, function* (error, res) {
            if (error) {
                if (jlog.jlogEnable)
                    jlog.error("error:", res, error);
            }
            else {
                // 创建模型
                if (jlog.jlogEnable)
                    jlog.debug("loadModel:" + res.fileMainName);
                // 创建对象
                var curObj1 = self.makeDefineByMod(res.obj);
                // 同步到库
                if (createTable) {
                    var timeSpTable = self.config.timeSpTable;
                    var timeSpField = self.config.timeSpField;
                    // 创建双表
                    if (timeSpTable) {
                        if (curObj1.tableName.substr(-4) === '_xjz') {
                            // 创建对象
                            var curObj2 = self.makeDefineByMod(res.obj, function (inObj) {
                                // 切换为临时表
                                inObj.tableName = inObj.tableName.replace(/_xjz/g, "_tmp");
                                inObj.opt.tableName = inObj.tableName;
                                // 检查字段
                                if (!inObj.attObj[timeSpField]) {
                                    if (jlog.jlogEnable)
                                        jlog.error("loss field [" + inObj.tableName + "]timeSpField:" + timeSpField);
                                    throw JERR.newError(JERR.PARAM_ERROR, "loss field [" + inObj.tableName + "]timeSpField:" + timeSpField);
                                }
                                else if (inObj.attObj[timeSpField].allowNull || inObj.attObj[timeSpField].allowNull === undefined) {
                                    if (jlog.jlogEnable)
                                        jlog.error("can't allowNull field [" + inObj.tableName + "]timeSpField:" + timeSpField);
                                    throw JERR.newError(JERR.PARAM_ERROR, "can't allowNull field [" + inObj.tableName + "]timeSpField:" + timeSpField);
                                }
                                // 检查索引
                                if (inObj.opt.indexes && inObj.opt.indexes.length) {
                                    var index = 0;
                                    for (; index < inObj.opt.indexes.length; index++) {
                                        const element = inObj.opt.indexes[index];
                                        if (element.fields && element.fields[0] === timeSpField)
                                            break;
                                    }
                                    if (index >= inObj.opt.indexes.length) {
                                        if (jlog.jlogEnable)
                                            jlog.error("loss index [" + inObj.tableName + "]timeSpField:" + timeSpField);
                                        throw JERR.newError(JERR.PARAM_ERROR, "loss index [" + inObj.tableName + "]timeSpField:" + timeSpField);
                                    }
                                }
                                else {
                                    if (jlog.jlogEnable)
                                        jlog.error("loss index [" + inObj.tableName + "]timeSpField:" + timeSpField);
                                    throw JERR.newError(JERR.PARAM_ERROR, "loss index [" + inObj.tableName + "]timeSpField:" + timeSpField);
                                }
                            });
                            yield self.createByDefine(curObj2);
                        }
                    }
                    // 创建表
                    yield self.createByDefine(curObj1);
                }
            }
        });
    }
    /**
     * 类型(长整形8字节20位)
     */
    BIGINT() {
        return "BIGINT,";
    }
    /**
     * 类型(整形4字节,10位)
     */
    INTEGER() {
        return "INTEGER,";
    }
    /**
     * 类型(日期时间型)
     */
    DATE() {
        return "DATE,";
    }
    /**
     * 类型(字符串)
     */
    STRING(len) {
        return "STRING," + (len ? len : 100);
    }
};
