
/*

// 文档
http://docs.sequelizejs.com/

// 普通查询
var a = yield mydb.q("select id from user where host=:name", { name: "张三" });

// 事务查询
yield mydb.t(function* (t,q) {
    var b = yield q("select 1");
    var c = yield q("select id from user where host=:name", { name: "张三" });
});

*/

import jbase_utils = require('./jbase_utils');
import jtype_utils = require('./jtype_utils');
import jret_utils = require('./jret_utils');
import jco = require("./jco");
import jmod_loader = require("./jmod_loader");

import Jlog = require('./jlog');
import JERR = require('./jerror_type');
var jlog = Jlog("JaDbSequelize");

var Sequelize = require('sequelize');

// /**
//  * 数据库选项
//  */
// interface DbOption extends Sequelize.Options {
//     /**
//      * 用户名
//      */
//     user: string,
//     /**
//      * 密码
//      */
//     password?: string,
//     /**
//      * 数据库
//      */
//     database?: string,
//     /**
//      * 提供更完整的错误信息(会损耗一点性能)
//      */
//     errInfo?: boolean,
//     /**
//      * 时间格式化,0=没有特殊处理,1=返回格林int戳,2="YYYY-MM-DD HH:mm:ss",3=IOS格式,"字符串自定义时间"=自定义(需要安装moment)
//      */
//     timeFmt?: number,
//     /**
//      * 启用时间双表,存在为清理周期null,单位(天)
//      */
//     timeSpTable?: number,
//     /**
//      * 双表时间字段名
//      */
//     timeSpField?: string
// }

/**
 * AndWhere类型
 */
type TypeAndWhere = {
    field: string,
    op: ">" | "<" | ">=" | "<=" | "=" | "!=" | "like" | "is null" | "is not null",
    value?: string | number
}

/**
 * 数据库表模型定义参数
 */
type TypeDbModDef = {
    tbName: string,
    attObj: any,
    opt: any
}

/**
 * 表回调定义
 */
type TypeTableDefineCbFun = (inObj: TypeDbModDef) => any

/**
 * 数据库查询参数
 */
type DbQueryParam = {
    /**
     * 查询语句,如果主参数存在将无效
     */
    sql?: string,
    /**
     * 事物体函数,若干存在则将在事务中执行
     */
    transaction?: any,
    /**
     * 时间字段,包含的列在执行后将自动转换为相应格式
     */
    timeField?: string[],
    /**
     * 时间字段格式,0=没有特殊处理,1=返回格林int戳,2="YYYY-MM-DD HH:mm:ss",3=IOS格式,"字符串自定义时间"=自定义(需要安装moment)
     */
    timeFmt?: number | string,
    /**
     * 分页查询选项
     */
    pageOpt?: {
        /**
         * 页大小
         */
        pageSize: number,
        /**
         * 当前页(1~N)
         */
        current?: number,
        /**
         * 需要计算总数
         */
        needTotal?: boolean
    },
    /**
     * order条件,要求查询里不能存在order
     */
    order?: string,
    /**
     * 与where条件,可以多组
     * sql:select * from xxx where a=123
     * andWhere:[{field:"b",op:"=",value:456}]
     * 组合后:select * from xxx where a=123 and (b=456)
     */
    andWhere?: TypeAndWhere[],
    /**
     * 该查询不使用时间双表
     */
    noUseTimeSpTable?: boolean,
    /**
     * 时间戳字段示例
     */
    dateTime?: string | Date,
    /**
     * 开始时间字段示例
     */
    beginTime?: string | Date,
    /**
     * 结束时间字段示例
     */
    endTime?: string | Date,
    /**
     * 映射字段值
     */
    [key: string]: any
};



/**
 * 数据库查询行回调
 */
type DbQueryCbFun = (row: { [key: string]: any }, index: number, all_rows: { [key: string]: any }[]) => any;

/**
 * 查询函数定义 
 */
type DbQueryDef = (sql_opt: string | DbQueryParam, sqlParam?: DbQueryParam, rowfun?: DbQueryCbFun) => any;

/**
 * 事务函数定义
 */
type DbTransactionDef = (t: any, q: any) => any;

/**
 * 获取表字符串
 * @param sqlStr SQL字段
 */
function getLastTabStr(sqlStr: string) {
    for (var index = sqlStr.length - 1; index >= 0; index--) {
        const ch = sqlStr.charCodeAt(index);
        if (ch === 45 || ch == 95 || (ch >= 48 && ch <= 57) || (ch >= 65 && ch <= 90) || (ch >= 97 && ch <= 122)) {
            continue;
        }
        else {
            return sqlStr.substr(index + 1);
        }
    }

    return sqlStr;
}


/**
 * MQ工具类
 */
export = class JaDbSequelize {

    /**
     * 自动载入的模型
     */
    public models: any = {};

    /**
     * 开启查询错误堆栈
     */
    public errInfo = false;

    /**
     * 时间格式化,0=默认时间类,1="YYYY-MM-DD HH:mm:ss",2=UTC整形数字,3=IOS格式,"字符串自定义时间"=自定义
     */
    public timeFmt = 0;


    /**
     * 库原型
     */
    public Sequelize = Sequelize;

    /**
     * 库对象
     */
    public db = <any>{};

    /**
     * 定义对象
     */
    public define = this.db.define;


    /**
     * 内部对象
     */
    protected innerObj = {
        /**
         * 启用时间双表,存在为清理周期null,单位(天)
         */
        timeSpTable: 0,
        /**
         * 清理表时间字段
         */
        timeSpField: "",
        /**
         * 时间分表清理对象
         */
        tSpTables: <{
            [key: string]: {
                /**
                 * 最后清理时间
                 */
                lastClean: number
            }
        }>{}
    };


    // 构造
    constructor() {

    }

    /**
     * 初始化
     * @param userCfg 配置
     * @returns 未定义
     */
    init(userCfg: any) {
        var self = this;
        var iobj = this.innerObj;
        return jco(function* (): any { // Begin Promise

            // 默认配置
            var config = <any>{
                "database": "postgres", // 数据库
                "user": "postgres", // 用户名
                "password": "", // 密码
                "host": "127.0.0.1", // 主机
                "port": 5432, // 端口
                "dialect": "postgres", // mysql,sqlite,postgres,mssql
                "pool": {
                    "max": 50, // 最大连接数
                    "min": 5, // 最小连接数
                    "acquire": 40000, // 连接数据库超时(ms)
                    "idle": 35000 // 断开连接空闲时间(ms)
                },
                // SQLite文件(不用可以不填)
                "storage": "path/to/database.sqlite",
                // http:// docs.sequelizejs.com/manual/tutorial/querying.html#operators
                "operatorsAliases": false, // 其他配置

                "timezone": jbase_utils.getTimeZoneStr(), // 时区

                "timeFmt": 0, // 时间格式化,0=默认时间类,1="YYYY-MM-DD HH:mm:ss",2=UTC整形数字,3=IOS格式,"字符串自定义时间"=自定义
                "errInfo": true, // 记录详细错误信息
                "logging": true, // 打印执行语句
                "timeSpTable": 0, // 拆分表时间(天数),0=不拆分
                "timeSpField": "jtime" // 时间拆分字段名
            }

            // 合并配置
            jbase_utils.unionObj(config, userCfg);

            // 日志记录
            if (config.logging === true || (typeof (config.logging) === 'number' && config.logging)) {
                config.logging = function (querySql: any) {
                    jlog.debug(querySql);
                }
            }

            // 参数检查
            if (config.dialect === "mysql2") config.dialect = "mysql";
            if (config.operatorsAliases === undefined) {
                config.operatorsAliases = false;
            }
            if (!config.database) config.database = "";
            if (!config.password) config.password = "";
            iobj.timeSpTable = config.timeSpTable ? config.timeSpTable : 0;
            iobj.timeSpField = config.timeSpField ? config.timeSpField : "jtime";


            // 创建数据库类
            if (jlog.jlogEnable) jlog.info("Load Database:dialect=" + config.dialect + ",database=" + config.database + ",user=" + config.user);
            var cur_db = new Sequelize(config.database, config.user, config.password, config);

            // 测试数据库连接
            while (true) {
                try {
                    yield cur_db.query('SELECT 1', { type: Sequelize.QueryTypes.SELECT });
                    if (jlog.jlogEnable) jlog.info("Test Database[" + config.database + "] Success");
                    break;
                } catch (error) {
                    if (jlog.jlogEnable) jlog.error("Test Database[" + config.database + "] Error:", error);
                    yield jbase_utils.sleep(5000);
                }
            }

            // 成功,功能重定向
            self.db = cur_db;
            self.define = self.db.define;
            self.errInfo = config.errInfo ? config.errInfo : false;
            self.timeFmt = config.timeFmt ? config.timeFmt : 0;
        });
    }

    /**
     * 数据库事务
     * @param tranFun 事务函数
     * @returns 无
     */
    t(tranFun: DbTransactionDef): Promise<any> {
        var self = this;
        var tranRet = self.db.transaction((transaction: any) => {
            return jco(tranFun(transaction, function (sql_opt: any, sqlParam: any, rowfun: any) {
                // 转换为事物查询,末尾带入事物函数
                return (<any>self).q(sql_opt, sqlParam, rowfun, transaction);
            }));
        });
        return <any>tranRet;
    }


    /**
     * 查询数据库
     * @param sql_opt SQL语句或param参数
     * @param sqlParam 参数,如果sql_opt也有参数,则两者将会合并
     * {
     * f1:123,f2:"str",
     * t:事务体(或者用transaction,省略不使用事务),
     * timeField:[时间列],
     * timeFmt:时间格式化(默认使用配置文件的timeFmt)
     * 不填=没有特殊处理,1=返回格林int戳,2="YYYY-MM-DD HH:mm:ss",3=IOS格式,"字符串自定义时间"=自定义(需要安装moment)
     * }
     * @param rowfun 处理函数fuction(row,index,all_rows)
     * @returns 查询对象
     */
    q(sql_opt: string | DbQueryParam, sqlParam?: DbQueryParam, rowfun?: DbQueryCbFun) {
        var self = this;
        var iobj = this.innerObj;

        // 获取快照
        var lastStack: any = {};
        if (self.errInfo) {
            Error.captureStackTrace(lastStack);
        }

        if (!self.db) {
            return Promise.reject(new Error("no init!"));
        }

        return jco(function* (): any {

            // 参数处理
            var sql: string;
            var param: DbQueryParam;

            if (typeof (sql_opt) === "string") {
                // 独立字符串参数
                sql = sql_opt;
                param = sqlParam ? sqlParam : {};
            }
            else if (sqlParam) {
                // 联合参数
                sql = sql_opt.sql ? sql_opt.sql : "";
                param = {};
                jbase_utils.unionObj(param, sql_opt, 0);
                jbase_utils.unionObj(param, sqlParam, 0);
            }
            else {
                // 独立对象参数
                sql = sql_opt.sql ? sql_opt.sql : "";
                param = sql_opt;
            }

            var q_op: any = {};

            // 展开SQL
            var lineSql = sql.replace(/(\r|\n)/g, " ").replace(/\(/g, " ( ").replace(/\)/g, " ) ");
            lastStack.lineSql = lineSql;

            // 操作方式
            var sql_op = lineSql.substr(0, 8);
            if (sql_op.indexOf("ins") >= 0) q_op.type = self.db.QueryTypes.INSERT;
            else if (sql_op.indexOf("del") >= 0) q_op.type = self.db.QueryTypes.DELETE;
            else if (sql_op.indexOf("upd") >= 0) q_op.type = self.db.QueryTypes.UPDATE;
            else q_op.type = self.db.QueryTypes.SELECT;

            //时间格式化
            var timeFmt = param.timeFmt;
            var timeField = param.timeField;

            // 事务体
            if (param.transaction && param.transaction.commit) {
                // 优先主参数
                q_op.transaction = param.transaction;
            }
            else if (arguments[3]) {
                // 次参数补充
                q_op.transaction = arguments[3];
            }

            // 模型
            if (param.model) {
                q_op.model = param.model;
            }

            // 参数体
            q_op.replacements = param;
            if (!timeFmt) timeFmt = self.timeFmt;

            // 查询
            var ret = [];
            var total = -2;
            var pageSize = 30;
            var pageNum = 0;
            try {
                // andWhere
                if (param.andWhere && param.andWhere.length) {
                    var sqlAndWhere = "";
                    for (var index = 0; index < param.andWhere.length; index++) {
                        const element = param.andWhere[index];
                        var covStr = (typeof (element.value) === "string") ? "'" : "";
                        if (sqlAndWhere.length > 0) sqlAndWhere += " and "
                        switch (element.op) {
                            case ">":
                            case "<":
                            case ">=":
                            case "<=":
                            case "=":
                            case "!=":
                            case "like":
                                sqlAndWhere += "(" + jbase_utils.filterSQL(element.field) + " " + element.op + " " + covStr + jbase_utils.filterSQL(element.value + "") + covStr + ")";
                                break;
                            case "is null":
                            case "is not null":
                                sqlAndWhere += "(" + jbase_utils.filterSQL(element.field) + " " + element.op + " )";
                                break;
                            default:
                        }
                    }
                    // 合并语句
                    if (sqlAndWhere) {
                        var sqlLc = lineSql.toLowerCase();
                        if (sqlLc.indexOf(" where ") > 0) {
                            lineSql += " and ( " + sqlAndWhere + " ) ";
                        }
                        else {
                            lineSql += " where " + sqlAndWhere;
                        }
                    }
                }

                // 时间分表
                if (!param.noUseTimeSpTable && iobj.timeSpTable) {
                    // 分析是否存在
                    var sqlSp = lineSql.split('_xjz');
                    if (sqlSp.length > 1) {
                        // 当前时间
                        var curIntTime = Date.now();
                        var cutTime = curIntTime - (iobj.timeSpTable * 24 * 60 * 60 * 1000);

                        // 提取表名称
                        for (var index = 0; index < sqlSp.length - 1; index++) {
                            var tbName = getLastTabStr(sqlSp[index]);
                            if (!tbName) continue;

                            // 没有则新建
                            if (!iobj.tSpTables[tbName]) {
                                iobj.tSpTables[tbName] = {
                                    lastClean: 0 // 立即开始
                                }
                            }

                            // 清理
                            if ((curIntTime - iobj.tSpTables[tbName].lastClean) > (24 * 60 * 60 * 1000)) {
                                iobj.tSpTables[tbName].lastClean = curIntTime;
                                jco(function* (): any {
                                    yield jbase_utils.sleep(30 * 1000);
                                    if (jlog.jlogEnable) jlog.info(`db clean table=${tbName}_tmp`);
                                    try {
                                        var delTime = new Date(curIntTime - ((iobj.timeSpTable + 1) * 24 * 60 * 60 * 1000));
                                        yield self.q(`delete from ${tbName}_tmp where ${iobj.timeSpField} < :mydate`, {
                                            noUseTimeSpTable: true,
                                            mydate: delTime
                                        });
                                        if (jlog.jlogEnable) jlog.info(`db clean table=${tbName}_tmp finish`);

                                    } catch (error) {
                                        if (jlog.jlogEnable) jlog.error(`db clean table=${tbName}_tmp`);
                                    }
                                });
                            }

                        }

                        // 查询类,根据时间字段选择表
                        if (q_op.type === self.db.QueryTypes.SELECT) {
                            var dateTime = jtype_utils.getDate(param.dateTime, cutTime, 1);
                            var beginTime = jtype_utils.getDate(param.beginTime, cutTime, 1);
                            var endTime = jtype_utils.getDate(param.endTime, cutTime, 1);

                            // 时间段有效
                            if ((dateTime >= cutTime) || (beginTime >= cutTime)) {
                                // 使用快速表
                                lineSql = lineSql.replace(/_xjz/g, "_tmp");
                            }
                            else if ((dateTime === cutTime) && (beginTime === cutTime) && (endTime === cutTime)) {
                                // 无任何限制条件
                                throw Error("Time Split Table param no [dateTime/beginTime/endTime]");
                            }
                        }
                        // 改写,双表操作
                        else {
                            ret = yield self.db.query(lineSql.replace(/_xjz/g, "_tmp"), q_op);
                        }
                    }
                }

                // 分页总数查询
                if (param.pageOpt && q_op.type === self.db.QueryTypes.SELECT) {
                    total = -1;
                    var fromIndex = lineSql.toLowerCase().indexOf(" from ");
                    if (fromIndex < 0) {
                        throw new Error("pageOpt err,nofind:from,sql=" + lineSql);
                    }

                    // 查询总数
                    if (param.pageOpt.needTotal && !param.pageOpt.current) {
                        var retCountList = yield self.db.query(`select count(1) as "count" ` + lineSql.substr(fromIndex), q_op);
                        if (retCountList.length > 0) {
                            total = parseInt(retCountList[0].count, 10);
                        }
                    }
                }

                // 加入order
                if (param.order && typeof (param.order) === "string") {
                    lineSql = lineSql + " order by " + jbase_utils.filterSQL(param.order);
                }

                // 组合分页语句
                if (param.pageOpt && q_op.type === self.db.QueryTypes.SELECT) {
                    pageSize = jtype_utils.getInt(param.pageOpt.pageSize, 30);
                    pageNum = jtype_utils.getInt(param.pageOpt.current, 1);
                    if (pageNum < 1) pageNum = 1;
                    var curNum = (pageNum - 1) * pageSize;
                    lineSql = `${lineSql} limit ${pageSize} offset ${curNum} `;
                }

                // 主查询
                ret = yield self.db.query(lineSql, q_op);

                // 记录处理
                if (!q_op.model && ret && ret.length > 0) { // 没有模型,有返回值

                    // 时间过滤
                    if (timeFmt && !timeField) {
                        for (var obj_name in ret[0]) {
                            if (ret[0][obj_name] instanceof Date) {
                                if (!timeField) timeField = []; // 创建时间数组
                                timeField.push(obj_name); // 放入时间字段
                            }
                        }
                    }

                    // 开始处理
                    if (rowfun || (timeField && timeFmt)) {
                        var i, j;
                        var isYield = jbase_utils.isYieldFun(rowfun);
                        for (i = 0; i < ret.length; i++) {
                            var cur_t = ret[i];

                            // 处理数据
                            if (rowfun) {
                                if (isYield) {
                                    yield rowfun(cur_t, i, ret);
                                }
                                else {
                                    rowfun(cur_t, i, ret);
                                }
                            }

                            // 处理时间字段
                            if (timeField && timeFmt) {
                                for (j = 0; j < timeField.length; j++) {
                                    var curTimVal = cur_t[timeField[j]];
                                    if (curTimVal) {
                                        cur_t[timeField[j]] = jtype_utils.getDate(curTimVal, curTimVal, timeFmt);
                                    }
                                }
                            }
                        }
                    }

                } // end 记录处理

                // 分页返回
                if (total > -2) {
                    // 添加分页信息
                    var pageRet = {
                        list: ret
                    }
                    jret_utils.addPageInfo(pageRet, total, pageNum, pageSize);
                    return pageRet;
                }
                else {
                    // 直接返回
                    return ret;
                }
            } catch (error: any) {
                // 额外的错误信息
                if (lastStack && lastStack.stack) {
                    error.lastStack = lastStack.stack;
                    error.lineSql = lastStack.lineSql
                }

                // 错误输出
                if (jlog.jlogEnable) jlog.error("error:", error);

                throw error;
            }

        });

    }

    /**
     * 通过模型创造表定义
     * @param dbModObj 模型对象
     * @param optCbFun 选项回调函数
     */
    makeDefineByMod(dbModObj: any, optCbFun?: TypeTableDefineCbFun) {
        var self = this;
        // 创建对象
        var curObj1 = dbModObj({
            define: (tbName: string, attObj: any, opt: any) => {
                if (!opt) opt = {};

                // 自定义回调处理
                if (optCbFun) {
                    var cbObj = {
                        tbName,
                        attObj,
                        opt
                    };
                    optCbFun(cbObj);
                    tbName = cbObj.tbName;
                }

                // 表明同创建名
                opt.tableName = tbName;
                if (!opt.timestamps) opt.timestamps = false;
                return self.db.define(tbName, attObj, opt);
            }
        }, Sequelize);

        // 放入模型表
        self.models[curObj1.name] = curObj1;

        return curObj1;
    }

    /**
     * 通过定义创建表,若存在则不创建
     * @param defineObj 对象定义
     */
    createByDefine(defineObj: any) {
        var self = this;
        return jco(function* (): any {

            var tSw = jlog.jlogEnable; // 临时关闭日志

            // 检查是否存在表
            try {
                jlog.jlogEnable = false;
                yield self.q('select 1 from ' + defineObj.tableName, { noUseTimeSpTable: true });
                jlog.jlogEnable = tSw; // 恢复日志
            }
            catch (err0) {
                jlog.jlogEnable = tSw; // 恢复日志

                // 创建表
                if (jlog.jlogEnable) jlog.info("create table:" + defineObj.tableName);
                yield defineObj.sync({});

                // 执行表数据
                var dataSql = jtype_utils.getArray(defineObj.options.dataSql, null, 1);
                if (dataSql) {
                    try {
                        if (jlog.jlogEnable) jlog.info("insert data to table:" + defineObj.tableName + ",count=" + dataSql.length);
                        for (var index = 0; index < dataSql.length; index++) {
                            var curDate = dataSql[index];
                            // 拼装语句
                            var sql_t1 = "";
                            for (const key in curDate) {
                                sql_t1 = sql_t1 + (sql_t1 ? ",:" : ":") + key;
                            }
                            // 执行
                            yield self.q({
                                sql: `insert into ${defineObj.tableName} (${sql_t1.replace(/:/g, "")})values(${sql_t1})`,
                                noUseTimeSpTable: true
                            }, curDate);
                        }
                    } catch (err1) {
                        if (jlog.jlogEnable) jlog.error("dataSql [" + defineObj.tableName + "] Error:", err1);
                        throw err1;
                    }
                }

            }
        });

    }

    /**
     * 载入模型,将会在models产生模型对象
     * @param modelDir 模型目录
     * @param createTable 自动创建表
     * @returns 模型原始定义对象
     */
    loadModel(modelDir: string, createTable?: boolean) {
        var self = this;
        var iobj = this.innerObj;
        return jmod_loader.loadAllModules(modelDir, false, function* (error, res) {
            if (error) {
                if (jlog.jlogEnable) jlog.error("error:", res, error);
            }
            else {
                // 创建模型
                if (jlog.jlogEnable) jlog.debug("loadModel:" + res.fileMainName);

                // 创建对象
                var curObj1 = self.makeDefineByMod(res.obj)

                // 同步到库
                if (createTable) {

                    // 创建双表
                    if (iobj.timeSpTable) {
                        if (curObj1.tableName.substr(-4) === '_xjz') {

                            // 创建对象
                            var curObj2 = self.makeDefineByMod(res.obj, function (inObj: TypeDbModDef) {

                                // 切换为临时表
                                inObj.tbName = inObj.tbName.replace(/_xjz/g, "_tmp");

                                // 检查字段
                                if (!inObj.attObj[iobj.timeSpField]) {
                                    if (jlog.jlogEnable) jlog.error("loss field [" + inObj.tbName + "]timeSpField:" + iobj.timeSpField);
                                    throw JERR.newError(JERR.PARAM_ERROR, "loss field [" + inObj.tbName + "]timeSpField:" + iobj.timeSpField);
                                }

                                // 检查索引
                                if (inObj.opt.indexes && inObj.opt.indexes.length) {
                                    var index = 0
                                    for (; index < inObj.opt.indexes.length; index++) {
                                        const element = inObj.opt.indexes[index];
                                        if (element.fields && element.fields[0] === iobj.timeSpField) break;
                                    }
                                    if (index >= inObj.opt.indexes.length) {
                                        if (jlog.jlogEnable) jlog.error("loss index [" + inObj.tbName + "]timeSpField:" + iobj.timeSpField);
                                        throw JERR.newError(JERR.PARAM_ERROR, "loss index [" + inObj.tbName + "]timeSpField:" + iobj.timeSpField);
                                    }
                                }
                                else {
                                    if (jlog.jlogEnable) jlog.error("loss index [" + inObj.tbName + "]timeSpField:" + iobj.timeSpField);
                                    throw JERR.newError(JERR.PARAM_ERROR, "loss index [" + inObj.tbName + "]timeSpField:" + iobj.timeSpField);
                                }
                            });

                            yield self.createByDefine(curObj2);
                        }
                    }

                    // 创建表
                    yield self.createByDefine(curObj1);
                }
            }
        });
    }

}
