"use strict";
/*
  // 简单使用说明

  // 创建Redis
  var redis = new JaRedis();

  // 初始化
  redis.init({
    host: '127.0.0.1', // 主机IP
    port: 6379, // 端口
    count: 2, // 连接数,默认:2
    database: 0, // 数据库ID,0~15
    keyfirst: "", // key头字符串
    options: {
      auth_pass: 'password', // 密码
      connect_timeout: 10000 // 连接超时
    }
  });

  // 写入
  yield redis.set("aaa", "123");

  // 读出
  console.log(yield redis.get("aaa"));

*/
const jbase_utils = require("./jbase_utils");
const jtype_utils = require("./jtype_utils");
const Jlog = require("./jlog");
var jlog = Jlog("JaIoRedis");
var redis = require('ioredis');
module.exports = class JaIoRedis {
    // 构造
    constructor() {
        /**
         * 内部对象
         */
        this.innerObj = {
            curIndex: 0, // 当前使用的连接
            url: undefined, // URL连接方式
            links: [], // 连接数组
            // 用户配置变量
            host: '127.0.0.1', // 主机
            port: 6379, // 端口
            database: 0, // 数据库ID
            options: {
                connect_timeout: 999999999 // 故障重试最大时间(不断重试,最多实验多久放弃)
            },
            keyfirst: "", // key头字符串,用于切割redis应用
            // 获取新的redis连接
            newConnect: () => {
                var iobj = this.innerObj;
                return new Promise((resolve, reject) => {
                    var client;
                    // 创建客户端
                    if (iobj.url) {
                        client = new redis(iobj.url, iobj.options);
                    }
                    else {
                        client = new redis(iobj.port ? iobj.port : 6379, iobj.host, iobj.options);
                    }
                    // 监听错误,通常指的是重试次数用尽后进来的
                    client.on('error', function (err) {
                        if (jlog.jlogEnable)
                            jlog.error(err); // 错误
                        client.removeAllListeners('ready');
                        client.removeAllListeners('error');
                        return reject(err); // 失败
                    });
                    // 监听就绪
                    client.on('ready', function () {
                        // 选择数据库
                        client.select(iobj.database, function (err) {
                            if (err) {
                                if (jlog.jlogEnable)
                                    jlog.error(err);
                                client.removeAllListeners('ready');
                                client.removeAllListeners('error');
                                return reject(err); // 失败
                            }
                            // 成功
                            if (jlog.jlogEnable)
                                jlog.info("link success");
                            return resolve(client); // 成功
                        });
                    });
                }); // End Promise
            },
        };
    }
    // 创建连接池
    init(cfg_client) {
        var iobj = this.innerObj;
        return new Promise((resolve, reject) => {
            if (jtype_utils.getArray(cfg_client)) {
                iobj.links = cfg_client;
                resolve();
            }
            else {
                // 配置项
                var cfg = cfg_client;
                if (typeof cfg.host === 'string')
                    iobj.host = cfg.host;
                if (typeof cfg.port === 'number')
                    iobj.port = cfg.port;
                if (typeof cfg.database === 'number')
                    iobj.database = cfg.database;
                if (typeof cfg.keyfirst === 'string')
                    iobj.keyfirst = (cfg.keyfirst === "") ? "" : (cfg.keyfirst + ":");
                // 合并对象
                jbase_utils.unionObj(iobj.options, cfg);
                jbase_utils.unionObj(iobj.options, cfg.options);
                // 密码选项兼容
                if (iobj.options.auth_pass)
                    iobj.options.password = iobj.options.auth_pass;
                if (!iobj.options.password) {
                    delete iobj.options.password;
                }
                // url配置
                if (typeof cfg.url === 'string') {
                    iobj.url = cfg.url;
                    iobj.host = undefined;
                    iobj.port = undefined;
                }
                // 断线重连
                iobj.options.retryStrategy = function (opt) {
                    if (jlog.jlogEnable)
                        jlog.error("redis link error,reconnect.");
                    return 5000;
                };
                // 建立连接
                var i = 0;
                var links_pm = [];
                var count = cfg.count ? cfg.count : 2;
                for (i = 0; i < count; i++) {
                    links_pm[i] = iobj.newConnect();
                }
                Promise.all(links_pm).then((res) => {
                    iobj.links = res;
                    resolve();
                }).catch((error) => {
                    reject(error);
                });
            }
        }); // End Promise
    }
    // 获取连接
    getLink() {
        var iobj = this.innerObj;
        iobj.curIndex++;
        if (iobj.curIndex >= iobj.links.length)
            iobj.curIndex = 0;
        return iobj.links[iobj.curIndex];
    }
    // 键转换
    getKey(key) {
        if (key.substr(0, 1) === "/") {
            // 从根读
            return key.substr(1);
        }
        // 加入keyfirst
        return this.innerObj.keyfirst + key;
    }
    /**
     * 设置键值
     * @param key 键
     * @param val 值
     * @param expSec 不存在或0为不限期
     * @returns 返回"OK"正常
     */
    set(key, val, expSec) {
        var self = this;
        key = self.getKey(key); // key头字符串
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 设置值
            if (expSec) {
                // 带时间参数
                client.setex(key, expSec, val, function (error, res) {
                    return error ? reject(error) : resolve(res);
                });
            }
            else {
                // 不带时间参数
                client.set(key, val, function (error, res) {
                    return error ? reject(error) : resolve(res);
                });
            }
        }); // End Promise
    }
    /**
     * 取值
     * @param key 键
     * @returns 返回值
     */
    get(key) {
        var self = this;
        key = self.getKey(key); // key头字符串
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 取出值
            client.get(key, function (error, res) {
                return error ? reject(error) : resolve(res);
            });
        }); // End Promise
    }
    /**
     * 删除键
     * @param key 键
     * @returns 未定义
     */
    del(key) {
        var self = this;
        key = self.getKey(key); // key头字符串
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 删除
            client.del(key, function (error, res) {
                return error ? reject(error) : resolve(res);
            });
        }); // End Promise
    }
    // 检测值回调型,不带key前缀
    getLockCb(key, expSec, client, cb) {
        // // 是否带时间
        // if ((typeof expSec === 'number') && expSec > 0) {
        //   client.set(key, "1", "NX EX", expSec, cb); // 带时间锁,"NX", "EX" ???
        // }
        // else {
        //   client.set(key, "1", "NX", cb);
        // }
    }
    /**
     * 获取锁
     * @param key 键值
     * @param expSec 最长锁定时间(s)-1=永久锁,直到代码解除
     * @param waitTimeMs 获取锁等待时间(ms),-1=一直等,直到获得锁
     * @returns 返回undefined为失败
     */
    getLockWait(key, expSec, waitTimeMs) {
        // var self = this;
        // key = self.getKey(key); // key头字符串
        // if (typeof waitTimeMs !== 'number') waitTimeMs = -1;
        // return new Promise<string | undefined>((resolve, reject) => { // Begin Promise
        //   // 取出连接
        //   var client = self.getLink();
        //   // 尝试锁
        //   this.getLockCb(key, expSec, client, (error:any, res:any) => {
        //     if (error) return reject(error);
        //     if (res) {
        //       // 锁成功
        //       return resolve(res);
        //     }
        //     else if (waitTimeMs === 0) // 只检测一次
        //     {
        //       // 未获得锁
        //       return resolve(undefined);
        //     }
        //     // 锁被持有,等待锁
        //     var chkTimer = setInterval(() => {
        //       // 等待时间
        //       if (waitTimeMs !== -1) {
        //         if (waitTimeMs <= 0) {
        //           // 未获得锁
        //           clearInterval(chkTimer);
        //           return resolve(undefined);
        //         }
        //         // 时间递减
        //         waitTimeMs -= 10;
        //       }
        //       // 判断锁
        //       this.getLockCb(key, expSec, client, (error:any, res:any) => {
        //         if (error) {
        //           clearInterval(chkTimer);
        //           return reject(error);
        //         }
        //         else if (res) {
        //           // 锁成功
        //           clearInterval(chkTimer);
        //           return resolve(res);
        //         }
        //       });
        //     }, 10);
        //   });
        // }); // End Promise
    }
    /**
     * 设置键值过期
     * @param key 键
     * @param expSec // 无时间或0则为不限期
     * @returns 返回null为失败
     */
    expire(key, expSec) {
        var self = this;
        key = self.getKey(key); // key头字符串
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 设置值
            if (expSec) {
                // 带时间参数
                client.expire(key, expSec, function (error, res) {
                    return error ? reject(error) : resolve(res);
                });
            }
            else {
                // 不带时间参数
                client.persist(key, function (error, res) {
                    return error ? reject(error) : resolve(res);
                });
            }
        }); // End Promise
    }
    /**
     * 取出过期时间
     * @param key 键值
     * @returns 返回-2=不存在,-1=存在,永久过期时间,>=0表示剩余时间
     */
    ttl(key) {
        var self = this;
        key = self.getKey(key); // key头字符串
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 取出值
            client.ttl(key, function (error, res) {
                return error ? reject(error) : resolve(res);
            });
        }); // End Promise
    }
    /**
     * 值加减
     * @param key 键
     * @param addNum // 递增的数值,为负则是递减
     * @returns 返回操作之后的值(键不存在则默认为0的基础上操作)
     */
    incrby(key, addNum) {
        var self = this;
        key = self.getKey(key); // key头字符串
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 改变值
            client.incrby(key, addNum, function (error, res) {
                return error ? reject(error) : resolve(res);
            });
        }); // End Promise
    }
    /**
     * 设置Hash子键值
     * @param key 键
     * @param field 列
     * @param val 值
     * @returns OK=成功
     */
    hset(key, field, val) {
        var self = this;
        key = self.getKey(key); // key头字符串
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 设置值
            client.hset(key, field, val, function (error, res) {
                return error ? reject(error) : resolve(res);
            });
        }); // End Promise
    }
    /**
     * 取出Hash子键值
     * @param key 键
     * @param field 列
     * @returns{string} 值
     */
    hget(key, field) {
        var self = this;
        key = self.getKey(key); // key头字符串
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 取出值
            client.hget(key, field, function (error, res) {
                return error ? reject(error) : resolve(res);
            });
        }); // End Promise
    }
    /**
     * 删除Hash子键
     * @param key 键
     * @param field 列
     * @returns OK=成功
     */
    hdel(key, field) {
        // var self = this;
        // key = self.getKey(key); // key头字符串
        // return new Promise<number>((resolve, reject) => { // Begin Promise
        //   // 取出连接
        //   var client = self.getLink();
        //   // 删除
        //   client.hdel(key, field, function (error:any, res:any) {
        //     return error ? reject(error) : resolve(res);
        //   });
        // }); // End Promise
    }
    /**
     * 取出所有hash值
     * @param key 键
     * @returns 返回对象
     */
    hgetall(key) {
        var self = this;
        key = self.getKey(key); // key头字符串
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 取出值
            client.hgetall(key, function (error, res) {
                return error ? reject(error) : resolve(res);
            });
        }); // End Promise
    }
    /**
     * Hash值加减
     * @param key 键
     * @param field 列
     * @param addNum // 递增的数值,为负则是递减
     * @returns{number} 返回操作之后的值(键不存在则默认为0的基础上操作)
     */
    hincrby(key, field, addNum) {
        var self = this;
        key = self.getKey(key); // key头字符串
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 改变值
            client.hincrby(key, field, addNum, function (error, res) {
                return error ? reject(error) : resolve(res);
            });
        }); // End Promise
    }
    /**
     * 队列左压入值
     * @param key 键
     * @param strArr 值,可以是字符串数组,按照顺序先入先出
     * @param maxCount 列表最大容量(可省),不填或0为无限大,超过maxCount+listKeepCount后删除listKeepCount个
     * @param listKeepCount 容量阈值(可省,与maxCount共存)
     * @returns 队列数量
     */
    llpush(key, strArr, maxCount, listKeepCount) {
        // var self = this;
        // key = self.getKey(key); // key头字符串
        // return new Promise<number>((resolve, reject) => { // Begin Promise
        //   // 取出连接
        //   var client = self.getLink();
        //   // 压入值
        //   client.lpush(key, strArr, function (error:any, res:any) {
        //     if (error) return reject(error);
        //     // 清理
        //     if (maxCount && (maxCount > 0)) {
        //       if (!listKeepCount) listKeepCount = 2;
        //       if (res > (maxCount + listKeepCount)) {
        //         // 缩减列表
        //         client.ltrim(key, 0, maxCount - 1, function (error:any, res:any) {
        //           return error ? reject(error) : resolve(maxCount);
        //         });
        //       }
        //       else {
        //         return resolve(res);
        //       }
        //     }
        //     else {
        //       return resolve(res);
        //     }
        //   });
        // }); // End Promise
    }
    /**
     * 队列右获取值,ack型
     * @param key 键
     * @param group 组(可省),用于多消费者(多消费者无法保证执行顺序)
     * @param waitTimeSec 等待时间(秒),-1=一直等,不填或0=立即返回,>0表示等待时间
     * @returns 值
     */
    lrget(key, group, waitTimeSec) {
        var self = this;
        key = self.getKey(key); // key头字符串
        if (!group) {
            group = "_nack";
        }
        else {
            group = "_nack_" + group;
        }
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 拿出备份表的数据
            client.lindex(key + group, 0, function (error, res) {
                if (error)
                    return reject(error);
                if (res)
                    return resolve(res);
                if (typeof waitTimeSec !== 'number')
                    waitTimeSec = 0;
                if (waitTimeSec === 0) {
                    // 不等待,直接返回
                    client.rpoplpush(key, key + group, function (error, res) {
                        return error ? reject(error) : resolve(res);
                    });
                }
                else {
                    // 等待返回
                    if (waitTimeSec < 0)
                        waitTimeSec = 0; // 不限时
                    client.brpoplpush(key, key + group, waitTimeSec, function (error, res) {
                        return error ? reject(error) : resolve(res);
                    });
                }
            });
        }); // End Promise
    }
    /**
     * 队列确认以前的获取
     * @param key 键
     * @param group 组(可省),用于多消费者
     * @returns OK
     */
    lack(key, group) {
        var self = this;
        key = self.getKey(key); // key头字符串
        if (!group) {
            group = "_nack";
        }
        else {
            group = "_nack_" + group;
        }
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 弹出值
            client.lpop(key + group, function (error, res) {
                return error ? reject(error) : resolve(res);
            });
        }); // End Promise
    }
    /**
     * 队列右获取值,无需ack(适合日志等,不依赖数据可靠性)
     * @param key 键
     * @param waitTimeSec 等待时间(秒),-1=一直等,不填或0=立即返回,>0表示等待时间
     * @returns 值
     */
    lrpop(key, waitTimeSec) {
        // var self = this;
        // key = self.getKey(key); // key头字符串
        // return new Promise<string | null>((resolve, reject) => { // Begin Promise
        //   // 取出连接
        //   var client = self.getLink();
        //   if (!waitTimeSec) {
        //     // 不等待,直接返回
        //     client.rpop(key, function (error:any, res:any) {
        //       return error ? reject(error) : resolve(res);
        //     });
        //   }
        //   else {
        //     // 等待返回
        //     if (waitTimeSec < 0) waitTimeSec = 0; // 不限时
        //     client.brpop(key, waitTimeSec, function (error:any, res:any) {
        //       return error ? reject(error) : resolve(res ? res[1] : null);
        //     });
        //   }
        // }); // End Promise
    }
    /**
     * 获取列表剩余元素数(不包括未确认的)
     * @param key 键
     * @returns 剩余元素数
     */
    llen(key) {
        var self = this;
        key = self.getKey(key); // key头字符串
        return new Promise((resolve, reject) => {
            // 取出连接
            var client = self.getLink();
            // 元素数量
            client.llen(key, function (error, res) {
                if (error)
                    return reject(error);
                return resolve(res);
            });
        }); // End Promise
    }
};
