
/*
  // 简单使用说明

  // 创建Redis
  var redis = new JaRedisLocal();

  // 初始化
  redis.init({
      "host": "./temp/redis.json", // 文件
      "saveTimeUp":100 // 存储间隔(s)
  });

  // 写入
  yield redis.set("aaa", "123");

  // 读出
  console.log(yield redis.get("aaa"));

*/

import jbase_utils = require('./jbase_utils');
import jco = require('./jco');
import jfile_utils = require('./jfile_utils');
import fs = require('fs');

import Jlog = require('./jlog');
var jlog = Jlog("JaRedisLocal");


/**
 * 本地Redis存储单元
 */
type TypeDataCell = {
  /**
   * 超时
   */
  t?: number,
  /**
   * 字符串内容
   */
  s?: string,
  /**
   * Hash
   */
  h?: { [Key: string]: string }
}

interface TypeInitOpt {
  /**
   * 保存时间间隔,省略后将不自动保存
   */
  saveTimeUp?: number,

  /**
   * 其他选项
   */
  [key: string]: any,
}

/**
 * Redis工具类
 */
export = class JaRedisLocal {

  /**
   * 内部对象
   */
  protected innerObj = {

    keyfirst: "", // key头字符串,用于切割redis应用

    saveTimeUp: 0, // 保存超时s

    dataObj: <{ [key: string]: TypeDataCell }>{}, // 数据对象

    change: <boolean>false, // 变化

    dataFile: "./redis_local.json" // 本地Redis数据文件

  };

  // 构造
  constructor() {

  }

  // 创建连接池
  init(cfg_client: TypeInitOpt) {
    var iobj = this.innerObj;

    // 文件位置
    if (typeof (cfg_client.host) === 'string') iobj.dataFile = cfg_client.host;
    if (typeof (cfg_client.saveTimeUp) === 'number') iobj.saveTimeUp = cfg_client.saveTimeUp;

    // 读取数据
    try {
      var DataStr = fs.readFileSync(iobj.dataFile, "utf8");
      if (DataStr) {
        if (jlog.jlogEnable) jlog.info("read redis data:" + iobj.dataFile);
        iobj.dataObj = JSON.parse(DataStr);
      }
    } catch (error) {
      iobj.dataObj = {};
    }
    if (!iobj.dataObj) iobj.dataObj = {};

    // 启动定时存储
    jco(function* (): any {
      var loopNum = 1;
      while (loopNum++) {
        // 延时
        yield jbase_utils.sleep(((iobj.saveTimeUp > 0) ? iobj.saveTimeUp : 300) * 1000);

        // 清理过期键值
        var nowTime = Date.now();
        var cleanCount = 0;
        //jlog.debug("clean local redis key");
        for (const key in iobj.dataObj) {
          var element = iobj.dataObj[key];
          if (element.t && element.t < nowTime) {
            delete iobj.dataObj[key];
            cleanCount++;
          }
        }
        if (cleanCount) {
          if (jlog.jlogEnable) jlog.debug(`clean local redis key count=${cleanCount},time=${Date.now() - nowTime}ms`);
        }

        // 存储
        if (iobj.change && iobj.saveTimeUp) {
          iobj.change = false;
          if (jlog.jlogEnable) jlog.info("write redis data " + iobj.dataFile);
          yield jfile_utils.writeFile(iobj.dataFile, JSON.stringify(iobj.dataObj));
          if (jlog.jlogEnable) jlog.info("write redis data ok");
        }
      }

    });


    // 完成
    return Promise.resolve();
  }

  /**
   * 设置键值
   * @param key 键
   * @param val 值
   * @param expSec 不存在或0为不限期
   * @returns 返回"OK"正常
   */
  set(key: string, val: any, expSec?: number) {
    var iobj = this.innerObj;
    // 保存
    iobj.dataObj[key] = {
      t: (expSec ? (Date.now() + (expSec * 1000)) : 0),
      s: val + ""
    }
    iobj.change = true; // 有变动

    // 返回
    return Promise.resolve("OK");
  }

  /**
   * 取值
   * @param key 键
   * @returns 返回值,null=无
   */
  get(key: string): Promise<string> {
    var iobj = this.innerObj;

    // 默认返回值
    var ret = null;

    // 读取
    var data = iobj.dataObj[key];
    if (data) {
      // 超时处理
      if (data.t && data.t < Date.now()) {
        delete iobj.dataObj[key];
      }
      else if (data.s !== undefined) {
        ret = data.s;
      }
    }

    // 返回
    return Promise.resolve(<any>ret);
  }

  /**
   * 删除键
   * @param key 键
   * @returns 返回删除数量
   */
  del(key: string) {
    var iobj = this.innerObj;

    var ret = 0;

    // 直接删除
    if (iobj.dataObj[key]) {
      delete iobj.dataObj[key];
      iobj.change = true; // 有变动
      ret = 1;
    }

    // 返回
    return Promise.resolve(ret);
  }


  // 检测值回调型,不带key前缀
  getLockCb(key: string, expSec: number, cb: any) {
    var self = this;
    var iobj = this.innerObj;

    // 默认返回值
    var ret = null;

    // 读取
    var data = iobj.dataObj[key];
    if (data) {
      // 超时处理
      if (data.t && data.t < Date.now()) {
        delete iobj.dataObj[key];
      }
      else if (data.s !== undefined) {
        ret = data.s;
      }
    }
    if (ret) {
      // 锁失败
      cb(null, null);
    }
    else {
      // 锁
      self.set(key, "1", expSec).then((res) => {
        // 锁成功
        cb(null, "OK");
      });
    }
  }

  /**
   * 获取锁
   * @param key 键值
   * @param expSec 最长锁定时间(s)-1=永久锁,直到代码解除
   * @param waitTimeMs 获取锁等待时间(ms),-1=一直等,直到获得锁
   * @returns 返回undefined为失败
   */
  getLockWait(key: string, expSec: number, waitTimeMs: number) {
    var self = this;
    if (typeof waitTimeMs !== 'number') waitTimeMs = -1;

    return new Promise<string | undefined>((resolve, reject) => { // Begin Promise

      // 尝试锁
      this.getLockCb(key, expSec, (error: any, res: any) => {
        if (error) return reject(error);
        if (res) {
          // 锁成功
          return resolve(res);
        }
        else if (waitTimeMs === 0) // 只检测一次
        {
          // 未获得锁
          return resolve(undefined);
        }

        // 等待锁
        jco(function* (): any {
          while (true) {

            // 等待时间
            if (waitTimeMs !== -1) {
              // 超时
              if (waitTimeMs <= 0) return undefined;

              // 时间递减
              waitTimeMs -= 10;
            }

            // 判断锁
            var lockSt = yield new Promise((resolve2, reject2) => {
              self.getLockCb(key, expSec, (err2: any, res2: any) => {
                if (err2) {
                  // 错误
                  return reject2(err2);
                }
                else if (res2) {
                  // 拿到
                  return resolve2(res2);
                }
                else {
                  // 没拿到
                  return resolve2(undefined);
                }
              });
            });

            // 拿到锁返回
            if (lockSt) return lockSt

            // 延时
            yield jbase_utils.sleep(10);
          }
        }).then((res) => {
          resolve(res);
        }).catch((err) => {
          reject(err);
        });


      });

    }); // End Promise
  }

  /**
   * 设置键值过期
   * @param key 键
   * @param expSec // 无时间或0则为不限期
   * @returns 返回0为未找到
   */
  expire(key: string, expSec?: number) {
    var iobj = this.innerObj;

    // 默认返回值
    var ret = 0;

    // 读取
    var data = iobj.dataObj[key];
    if (data) {
      // 超时处理
      if (data.t && data.t < Date.now()) {
        delete iobj.dataObj[key];
      }
      else {
        data.t = (expSec ? (Date.now() + (expSec * 1000)) : 0);
        ret = 1;
      }
    }

    // 返回
    return Promise.resolve(ret);
  }


  /**
   * 取出过期时间
   * @param key 键值
   * @returns 返回-2=不存在,-1=存在,永久过期时间,>=0表示剩余时间
   */
  ttl(key: string) {
    return new Promise<number>((resolve, reject) => { // Begin Promise
      resolve(-2);
    });
  }

  /**
   * 值加减
   * @param key 键
   * @param addNum // 递增的数值,为负则是递减
   * @returns 返回操作之后的值(键不存在则默认为0的基础上操作)
   */
  incrby(key: string, addNum: number) {
    var self = this;
    var iobj = this.innerObj;

    // 默认返回值
    var ret = 0;

    // 读取
    var data = iobj.dataObj[key];
    if (data) {
      // 超时处理
      if (data.t && data.t < Date.now()) {
        delete iobj.dataObj[key];
      }
      else if (data.s) {
        ret = parseInt(data.s, 10);
      }
    }
    ret += addNum;

    // 设置并返回
    return self.set(key, ret).then(function () {
      iobj.change = true; // 有变动
      return ret;
    });
  }

  /**
   * 设置Hash子键值(key的超时时间不会被更改)
   * @param key 键
   * @param field 列
   * @param val 值
   * @returns 并且值设置成功，返回 1 如果哈希表中域 field 已经存在且旧值已被新值覆盖，返回 0 
   */
  hset(key: string, field: string, val: any) {
    var iobj = this.innerObj;

    var ret = 1;

    // 取出对象
    var data = iobj.dataObj[key];
    if (!data) {
      data = { h: {} };
      iobj.dataObj[key] = data;
    }

    // 保存
    if (!data.h) data.h = {};
    if (data.h[field] !== undefined) ret = 0;
    data.h[field] = val + "";
    iobj.change = true; // 有变动

    // 返回
    return Promise.resolve(ret);
  }

  /**
   * 取出Hash子键值
   * @param key 键
   * @param field 列
   * @returns{string} 值,null=无
   */
  hget(key: string, field: string): Promise<string> {
    var iobj = this.innerObj;

    // 默认返回值
    var ret = null;

    // 读取
    var data = iobj.dataObj[key];
    if (data) {
      // 超时处理
      if (data.t && data.t < Date.now()) {
        delete iobj.dataObj[key];
      }
      else if (data.h !== undefined) {
        ret = data.h[field];
      }
    }

    // 返回
    return Promise.resolve(<any>ret);
  }

  /**
   * 删除Hash子键
   * @param key 键
   * @param field 列
   * @returns 删除数量
   */
  hdel(key: string, field: string) {
    var iobj = this.innerObj;

    // 默认返回值
    var ret = 0;

    // 读取
    var data = iobj.dataObj[key];
    if (data) {
      // 超时处理
      if (data.t && data.t < Date.now()) {
        delete iobj.dataObj[key];
      }
      else if (data.h !== undefined) {
        if (data.h[field]) {
          iobj.change = true; // 有变动
          delete data.h[field];
          ret++;
        }
      }
    }

    // 返回
    return Promise.resolve(ret);
  }

  /**
   * 取出所有hash值
   * @param key 键
   * @returns 返回对象
   */
  hgetall(key: string) {
    var iobj = this.innerObj;

    // 默认返回值
    var ret = <{ [key: string]: string } | null>null;

    // 读取
    var data = iobj.dataObj[key];
    if (data) {
      // 超时处理
      if (data.t && data.t < Date.now()) {
        delete iobj.dataObj[key];
      }
      else if (data.h !== undefined) {
        ret = {};
        for (const key in data.h) {
          ret[key] = data.h[key];
        }
      }
    }

    // 返回
    return Promise.resolve(ret);
  }


  /**
   * Hash值加减
   * @param key 键
   * @param field 列
   * @param addNum // 递增的数值,为负则是递减
   * @returns{number} 返回操作之后的值(键不存在则默认为0的基础上操作)
   */
  hincrby(key: string, field: string, addNum: number) {
    return new Promise<number>((resolve, reject) => { // Begin Promise
      resolve(1);
    });
  }

  /**
   * 队列左压入值
   * @param key 键
   * @param strArr 值,可以是字符串数组,按照顺序先入先出
   * @param maxCount 列表最大容量(可省),不填或0为无限大,超过maxCount+listKeepCount后删除listKeepCount个
   * @param listKeepCount 容量阈值(可省,与maxCount共存)
   * @returns 队列数量
   */
  llpush(key: string, strArr: string | string[], maxCount?: number, listKeepCount?: number) {
    return new Promise<number>((resolve, reject) => { // Begin Promise
      resolve(1);
    });
  }


  /**
   * 队列右获取值,ack型
   * @param key 键
   * @param group 组(可省),用于多消费者(多消费者无法保证执行顺序)
   * @param waitTimeSec 等待时间(秒),-1=一直等,不填或0=立即返回,>0表示等待时间
   * @returns 值
   */
  lrget(key: string, group?: string | null, waitTimeSec?: number) {
    return new Promise<string | null>((resolve, reject) => { // Begin Promise
      resolve(null);
    });
  }

  /**
   * 队列确认以前的获取
   * @param key 键
   * @param group 组(可省),用于多消费者
   * @returns OK
   */
  lack(key: string, group?: string | null) {
    return new Promise<string | null>((resolve, reject) => { // Begin Promise
      resolve("OK");
    });
  }

  /**
   * 队列右获取值,无需ack(适合日志等,不依赖数据可靠性)
   * @param key 键
   * @param waitTimeSec 等待时间(秒),-1=一直等,不填或0=立即返回,>0表示等待时间
   * @returns 值
   */
  lrpop(key: string, waitTimeSec?: number) {
    return new Promise<string | null>((resolve, reject) => { // Begin Promise
      resolve(null);
    });
  }


  /**
   * 获取列表剩余元素数(不包括未确认的)
   * @param key 键
   * @returns 剩余元素数
   */
  llen(key: string) {
    return new Promise<number>((resolve, reject) => { // Begin Promise
      resolve(0);
    });
  }

}

