"use strict";
const jbase_utils = require("../jbase_utils");
const jdocode = require("../jdocode");
const jco = require("../jco");
const Jlog = require("../jlog");
var jlog = Jlog("JaPublisher");
var amqp = require('amqp');
module.exports = class JaPublisher {
    /**
     * 构造
     * @param channel mq通道
     */
    constructor(channel) {
        /**
         * 内部对象
         */
        this.innerObj = {
            /**
             * 信道
             */
            m_channel: undefined,
            /**
             * exchange
             */
            m_exchange: undefined,
            /**
             * queue
             */
            m_queue: undefined,
            /**
             * routeKey
             */
            m_routeKey: "",
            /**
             * 消息存活时间
             */
            m_ttl: null,
            /**
             * exchange名称
             */
            m_exName: "",
            /**
             * 配置
             */
            config: {
                exchange: {
                    name: "amq.topic"
                },
                publish: {
                    routing: "default",
                    confirm: true, // 发送者需要得到服务器回执
                    msgTtl: null // 消息生存时间(秒),为空则永不过期
                }
            }
        };
        var iobj = this.innerObj;
        iobj.m_channel = channel;
    }
    /**
     * 初始化发布者
     * @param opt
        {
            exchange: { // Exchange
                name: 'DevTranService_Msg_v1', // Exchange名称
                
                // <可选>如果是被动接受者则不需下列对象
                options: { // 创建Exchange的属性
                    type: 'topic', // direct:完全匹配,topic:不完全匹配,fanout:不进行过滤
                    durable: true, // 开启持久化
                    autoDelete: false // 当删除了所有的queue时自动删除Exchange
                }
            },

            // 发布配置<可选>
            publish: {
                confirm: true, // 发送者需要得到服务器回执(默认:true)
                msgTtl:null // 消息生存时间(秒),为空则永不过期(默认:null)
            },

            // 为第三方应用创建通道,防止丢失<可选>
            queues: [
                // 与queue语法一致
            ]
            
        }
     * @returns 无
     */
    init(opt) {
        var iobj = this.innerObj;
        //检查参数
        if (!opt) {
            throw Error('Invalid options');
        }
        // 复制配置
        jbase_utils.unionObj(iobj.config, opt);
        if (iobj.config.exchange && iobj.config.publish) {
            var exCfg = iobj.config.exchange;
            var pubCfg = iobj.config.publish;
            // 默认路由名
            if (pubCfg.routing)
                iobj.m_routeKey = pubCfg.routing;
            // 消息存活时间
            iobj.m_ttl = pubCfg.msgTtl;
            // 交换器名
            iobj.m_exName = exCfg.name ? exCfg.name : "amq.topic";
            return jco(function* () {
                //检查参数
                if (!iobj.m_channel) {
                    throw Error('Invalid options');
                }
                // 创建Exchange
                if (exCfg.options) {
                    exCfg.options.confirm = pubCfg.confirm;
                }
                else {
                    // 使用现有的,需要等待出现
                    // 通过Queue进行检测
                    var tmpQueue = undefined;
                    yield new Promise((resolve, reject) => {
                        var queueMain = iobj.m_channel.queue(jdocode.getUUIDHexStr(), {}, () => {
                            queueMain.bind(iobj.m_exName, jdocode.getUUIDHexStr(), (queueBind) => {
                                tmpQueue = queueBind;
                                // 已决
                                resolve();
                            });
                        });
                    });
                    //销毁临时queue
                    if (tmpQueue)
                        tmpQueue.destroy();
                }
                // 创建或绑定Exchange
                iobj.m_exchange = yield new Promise((resolve, reject) => {
                    var retExchange = iobj.m_channel.exchange(iobj.m_exName, exCfg.options ? exCfg.options : { passive: true, confirm: pubCfg.confirm }, () => {
                        iobj.m_exchange = retExchange; // exchange赋值
                        resolve(retExchange);
                    });
                    retExchange.on("error", (err) => {
                        if (jlog.jlogEnable)
                            jlog.error(err);
                        reject(err);
                    });
                });
                // 创建Queue组,用于辅助第三方
                if (iobj.config.queues && iobj.config.queues.length) {
                    // 创建Queue
                    for (var index = 0; index < iobj.config.queues.length; index++) {
                        var curQuCfg = iobj.config.queues[index];
                        if (jlog.jlogEnable)
                            jlog.info("create queue:" + curQuCfg.name);
                        yield new Promise((resolve, reject) => {
                            // 队列处理
                            var retQueue = iobj.m_channel.queue(curQuCfg.name, curQuCfg.options ? curQuCfg.options : { passive: true }, // 创建或拿出队列
                            () => {
                                // 绑定Exchange
                                retQueue.bind(iobj.m_exName, curQuCfg.routing ? curQuCfg.routing : "default", (queue) => {
                                    resolve();
                                });
                            });
                            retQueue.on("error", (err) => {
                                if (jlog.jlogEnable)
                                    jlog.error(err);
                                reject(err);
                            });
                        });
                    }
                }
            });
        }
        else {
            // 错误
            return Promise.reject(new Error("config Error"));
        }
    }
    /**
     * 发布消息(原始)
     * @param routeKey 路由key
     * @param msg 消息
     * @param opt 选项
     * @return 无
     */
    publishEx(routeKey, msg, opt) {
        var iobj = this.innerObj;
        //检查参数
        if (!iobj.m_exchange) {
            return Promise.reject(Error('Invalid options'));
        }
        // 开始发送
        return new Promise((resolve, reject) => {
            try {
                // 发布
                iobj.m_exchange.publish(routeKey, msg, opt, (err, err_msg) => {
                    if (err === false) {
                        // 成功
                        return resolve();
                    }
                    // 错误
                    if (jlog.jlogEnable)
                        jlog.error(new Error(err_msg));
                    return reject(err_msg);
                });
            }
            catch (error) {
                if (jlog.jlogEnable)
                    jlog.error(error);
                return reject(error);
            }
        });
    }
    /**
    * 发布消息
    * @param msg 消息
    * @param routeKey 路由key(可省,没绑定Queue必须填写)
    * @param headers 消息头(可省,或者填入{})
    * @param options 选项(可省){expiration:"120000"超时}
    * @param autoTryCount 自动重试次数(可省,-1=无限循环)
    * @return 无
    */
    publishMsg(msg, routeKey, headers, options, autoTryCount) {
        var self = this;
        var iobj = this.innerObj;
        // 发布消息
        return jco(function* () {
            //检查参数
            if (!iobj.m_exchange) {
                throw Error('Invalid options');
            }
            // 参数补全
            if (!routeKey)
                routeKey = iobj.m_routeKey;
            // 选项
            var opt;
            if (options) {
                opt = {};
                jbase_utils.unionObj(opt, options);
            }
            else {
                opt = {
                    deliveryMode: 2
                };
            }
            if (headers)
                opt.headers = headers;
            // 设置消息超时
            if (opt.expiration) {
                opt.expiration += "";
            }
            else if (iobj.m_ttl) {
                opt.expiration = (iobj.m_ttl * 1000) + ""; //(new Date(Date.now() + (iobj.m_ttl * 1000))).toISOString();
            }
            // 发送重试
            if (!autoTryCount)
                autoTryCount = 0;
            if (autoTryCount < 0)
                autoTryCount = 999999999;
            for (; autoTryCount >= 0; autoTryCount--) {
                try {
                    // 发布
                    yield self.publishEx(routeKey, msg, opt);
                    break;
                }
                catch (error) {
                    yield jbase_utils.sleep(5000); // 错误重试
                }
            }
        });
    }
};
