"use strict";
const jbase_utils = require("../jbase_utils");
const jdocode = require("../jdocode");
const JERR_S = require("../jerror_type");
const jco = require("../jco");
const Jlog = require("../jlog");
var jlog = Jlog("JaSubscriber");
var amqp = require('amqp');
module.exports = class JaSubscriber {
    /**
     * 构造
     * @param channel mq通道
     */
    constructor(channel) {
        /**
         * 内部对象
         */
        this.innerObj = {
            /**
             * 信道
             */
            m_channel: undefined,
            /**
             * exchange
             */
            m_exchange: undefined,
            /**
             * queue
             */
            m_queue: undefined,
            /**
             * routeKey
             */
            m_routeKey: "",
            /**
             * uuid
             */
            uuidBase: jdocode.getUUID(8),
            /**
             * uuidIndex
             */
            uuidIndex: 0,
            /**
             * 自动确认
             */
            autoFunAck: false,
            /**
             * 回调函数是否为异步函数
             */
            isYieldCbFun: true,
            /**
             * 回调函数
             */
            cbFun: null,
            /**
             * 订阅回调函数
             */
            subCbFun: null,
            /**
             * 返回必须为JSON对象
             */
            isJsonObj: false,
            /**
             * 配置
             */
            config: {
                // 订阅配置
                subscribe: {
                    ack: true, // 需要应答(该选项为true时,prefetchCount,autoFunAck可用)
                    prefetchCount: 5, // 预读取数量,0=无限制
                    isJsonObj: false, // 解析为json对象体,若解析失败则消息为null
                    autoFunAck: true // function*结束后自动应答
                },
                // 默认绑定的Exchange
                exchange: {
                    name: 'amq.topic', // Exchange名称
                },
                // 通道配置
                queue: {
                    name: 'jqueue_default', // Queue名称
                    // 新建队列选项
                    options: {
                        durable: true, // 开启持久化
                        autoDelete: false // 当连接全部断开时自动删除Queue
                    }
                }
            },
            /**
             * 订阅处理函数
             */
            subCbProcess: (message, headers, deliveryInfo, messageObject) => {
                var iobj = this.innerObj;
                iobj.uuidIndex++;
                var curUUID = iobj.uuidBase + ("0000000000000000" + iobj.uuidIndex).substr(-16);
                // 取出消息体
                var inMsg = message;
                if (iobj.isJsonObj && message && (message.data instanceof Buffer)) {
                    try {
                        inMsg = JSON.parse(message.data.toString());
                        if (typeof (inMsg) !== "object") {
                            throw (JERR_S.newError(JERR_S.PARAM_ERROR));
                        }
                    }
                    catch (error) {
                        deliveryInfo.parseError = error;
                        deliveryInfo.rawData = message.data;
                        inMsg = null;
                        if (jlog.jlogEnable)
                            jlog.error(curUUID, 'msg: parse Error');
                    }
                }
                // 打上标
                if (headers)
                    headers.juuid = curUUID;
                if (jlog.jlogEnable)
                    jlog.debug(curUUID, 'msg get,key=' + deliveryInfo.routingKey + ",msg=", inMsg, ",header=", headers);
                if (iobj.isYieldCbFun) {
                    // 异步执行
                    jbase_utils.seqDo(iobj.uuidBase, function* () {
                        // 日志
                        if (jlog.jlogEnable)
                            jlog.debug(curUUID, 'msg seqDo');
                        // 执行回调
                        try {
                            yield iobj.cbFun(inMsg, headers, deliveryInfo, messageObject);
                        }
                        catch (error) {
                            if (jlog.jlogEnable)
                                jlog.error(curUUID, 'msg error:', error);
                        }
                        // 确认消息
                        if (iobj.autoFunAck) {
                            if (jlog.jlogEnable)
                                jlog.debug(curUUID, 'msg acknowledge');
                            messageObject.acknowledge(false);
                        }
                    });
                }
                else {
                    // 执行回调
                    try {
                        iobj.cbFun(inMsg, headers, deliveryInfo, messageObject);
                    }
                    catch (error) {
                        if (jlog.jlogEnable)
                            jlog.error(curUUID, 'msg error:', error);
                    }
                    // 确认消息
                    if (iobj.autoFunAck) {
                        if (jlog.jlogEnable)
                            jlog.debug(curUUID, 'msg acknowledge');
                        messageObject.acknowledge(false);
                    }
                }
            },
            /**
             * Queue绑定
             */
            qBind(queue, exchangeName, routingKey) {
                return new Promise((resolve, reject) => {
                    queue.bind(exchangeName, routingKey ? routingKey : "#", () => {
                        if (jlog.jlogEnable)
                            jlog.info(`qBind ${exchangeName} -> ${routingKey}`);
                        resolve();
                    });
                }); // End Promise
            },
            /**
             * 订阅
             */
            subscribe: (queue, options, cbFun) => {
                var iobj = this.innerObj;
                iobj.autoFunAck = options.autoFunAck ? true : false;
                iobj.isJsonObj = options.isJsonObj ? true : false;
                iobj.isYieldCbFun = jbase_utils.isYieldFun(cbFun);
                iobj.cbFun = cbFun;
                if (jlog.jlogEnable)
                    jlog.info("ready,autoFunAck=" + iobj.autoFunAck + ",isJsonObj=" + iobj.isJsonObj);
                // 订阅
                queue.subscribe(options, function (message, headers, deliveryInfo, messageObject) {
                    iobj.subCbProcess(message, headers, deliveryInfo, messageObject);
                });
            }
        };
        var iobj = this.innerObj;
        iobj.m_channel = channel;
    }
    /**
     * 初始化订阅者
     * @param opt
        {
            // 订阅配置
            subscribe: {
                ack: true, // 需要应答(该选项为true时,prefetchCount,autoFunAck可用)
                prefetchCount: 10, // 预读取数量,0=无限制
                isJsonObj: true, // 解析为json对象体,若解析失败则消息为null
                autoFunAck: true // function*结束后自动应答
            },

            // 默认绑定的Exchange
            exchange: {
                name: 'DevTranService_Recv_v1', // Exchange名称

                // 如果是被动接受者则不需下列对象
                options: { // 创建Exchange的属性
                    type: 'topic', // direct:完全匹配,topic:不完全匹配,fanout:不进行过滤
                    durable: true, // 开启持久化
                    autoDelete: false // 当删除了所有的queue时自动删除Exchange
                }
            },

            // 通道配置
            queue: {
                name: 'DevTranService_Recv_v1', // Queue名称

                // 默认绑定的routekey
                routing: 'api_v1',

                // <可省>绑定其他Exchange和routing
                bindEx:[
                    {
                        exchange:"xx",
                        routing:"yyyy"
                    }
                ],

                // <可省>如果是被动使用共享的Queue则不需下列对象
                options: {
                    durable: true, // 开启持久化
                    autoDelete: false // 当连接全部断开时自动删除Queue
                }
            }
        }
     * @param cb 消息回调函数
     * @returns 无
     */
    init(opt, cb) {
        var iobj = this.innerObj;
        //检查参数
        if (!opt) {
            throw Error('Invalid options');
        }
        // 复制配置
        jbase_utils.unionObj(iobj.config, opt);
        return jco(function* () {
            //检查参数
            if (!iobj.config.queue) {
                throw Error('Invalid options');
            }
            // 创建Exchange
            if (iobj.config.exchange && iobj.config.exchange.options) {
                iobj.m_exchange = yield new Promise((resolve, reject) => {
                    // 创建Exchange
                    var retExchange = iobj.m_channel.exchange(iobj.config.exchange.name, iobj.config.exchange.options, () => {
                        resolve(retExchange);
                    });
                    // 捕获错误
                    retExchange.on("error", (err) => {
                        if (jlog.jlogEnable)
                            jlog.error(err);
                        reject(err);
                    });
                });
            }
            // 创建Queue
            if (iobj.config.queue.options) {
                jlog.info("Queue:" + iobj.config.queue.name);
                yield new Promise((resolve, reject) => {
                    // 创建Queue(不存在或者订阅者主动)
                    var retQueue = iobj.m_channel.queue(iobj.config.queue.name, iobj.config.queue.options ? iobj.config.queue.options : { passive: true }, () => {
                        iobj.m_queue = retQueue;
                        // 完成
                        return resolve();
                    });
                    // 错误处理
                    retQueue.on("error", (err) => {
                        if (jlog.jlogEnable)
                            jlog.error(err);
                        reject(err);
                    });
                });
                // 绑定当前Queue
                if (iobj.config.queue.routing) {
                    yield iobj.qBind(iobj.m_queue, iobj.config.exchange.name, iobj.config.queue.routing);
                }
                // 绑定其他Queue
                var bindEx = iobj.config.queue.bindEx;
                if (bindEx) {
                    for (var index = 0; index < bindEx.length; index++) {
                        var element = bindEx[index];
                        yield iobj.qBind(iobj.m_queue, element.exchange ? element.exchange : iobj.config.exchange.name, element.routing);
                    }
                }
                // 开始订阅
                iobj.subCbFun = cb;
                iobj.subscribe(iobj.m_queue, iobj.config.subscribe, iobj.subCbFun);
            }
        });
    }
};
