'use strict';

var CO              = require('../jco');
var TypeUtils       = require('./TypeUtils');

class MQSubscriber {
    constructor(channel) {
        this.m_Channel      = channel;
        this.m_Exchange     = null;
        this.m_Queue        = null;
        this.m_TempQueue    = [];
        this.m_CB           = null;
        this.m_Idle         = true;

        /** Map Data */
        this.m_MapSet       = new Map();
        this.m_MaxConc      = 100;
        this.m_CurConc      = 0;
    }

    /**
     * 
     * @param {*} Options 
     * Option = {
     *  Queue : {
     *      name : 'xxx',                           // Queue的名字
     *      routing : 'xxx',                        // 路由Key,
     *      Options : {
     *          passive : true/false(default),          // true,只检查不创建
     *          durable : true/false(default),          // true,开启持久化
     *          exclusive   : true/false(default),      // true,私有队列
     *          autoDelete  : false/true(default),      // true,在没有消费者使用后，自动删除
     *          noDeclare   : true/false(default),      // true,用来删除同名的队列
     *          arguments   : map,                      // 
     *          closeChannelOnUnsubscribe   : true/false(default),  // true,在调用unsubscri之后，关闭Channel
     *      }
     *  },  // requierd
     *  Exchange : {
     *      name : 'xx',                                // Exchange的名字
     *      Options : {
     *          type : 'direct'/'fanout'/'topic'(default),  // exchange的类型
     *          passive : true/false(default),              // true,只检查不创建
     *          durable : true/false(default),              // true,开启持久化
     *          autoDelete  : false/true(default),          // true,在没有queues绑定之后，自动删除
     *          noDeclare   : true/false(default),          // true,用来删除同名的交换器
     *          confirm     : true/false(default),          // true,开启confirm模式（手动确认）
     *          arguments   : map,                          // 
     *      }   // undefined(default)
     *  }   // requierd
     * } 
     */
    Init(Options) {
        var self    = this;
        return CO(function *() {
            if (TypeUtils.isEmptyObj(self.m_Channel) === true) {
                throw 'Invalid channel';
            }
            if (TypeUtils.isEmptyObj(Options.Queue) === true
                || TypeUtils.isEmptyObj(Options.Exchange) === true
                || TypeUtils.isString(Options.Exchange.name) === false) {
                throw 'Invalid options';
            }

            if (TypeUtils.isEmptyObj(Options.Exchange.Options) === false) {
                self.m_Exchange = yield new Promise((resolve, reject) => {
                    self.m_Channel.exchange(Options.Exchange.name,
                        Options.Exchange.Options, (exchange) => {
                        resolve(exchange);
                    });
                });
            }

            self.m_Queue    = yield new Promise((resolve, reject) => {
                self.m_Channel.queue(Options.Queue.name,
                    Options.Queue.Options, (queue) => {
                    resolve(queue);
                })
            });
            yield new Promise((resolve, reject) => {
                self.m_Queue.bind(Options.Exchange.name, Options.Queue.routing, (queue) => {
                    resolve();
                })
            });
        })
    }

    Subscribe(CB, Options) {
        if (TypeUtils.isEmptyObj(this.m_Queue) === true) {
            throw 'No init';
        }

        var self    = this;
        self.m_Queue.subscribe(Options, (message, headers, deliveryInfo, messageObject) => {
            if (TypeUtils.isFunction(CB) === true) {
                CB(message, headers, deliveryInfo, messageObject)
            }
        });
    }

    SubscribeEx(CB, Options) {
        if (TypeUtils.isEmptyObj(this.m_Queue) === true) {
            throw 'No init';
        }

        this.m_CB   = CB;
        var Self = this;
        Self.m_Queue.subscribe(Options, (message, headers, deliveryInfo, messageObject) => {
            // console.log(this.m_TempQueue.length);
            Self.m_TempQueue.push({
                Message : message,
                Headers : headers,
                DeliveryInfo : deliveryInfo,
                MessageObject : messageObject
            });
            if (Self.m_Idle) {
                Self.m_Idle = false;
                setTimeout(() => {
                    Self._Next();
                }, 0);
            }
        });
    }

    SubscribeMap(CB, Options, MaxConc) {
        if (TypeUtils.isEmptyObj(this.m_Queue) === true) {
            throw 'No init';
        }

        this.m_CB       = CB;
        this.m_MaxConc  = parseInt(MaxConc) ? parseInt(MaxConc) : 100;
        this.m_CurConc  = 0;
        var Self        = this;
        
        Self.m_Queue.subscribe(Options, (message, headers, deliveryInfo, messageObject) => {
            var CurMsg = JSON.parse(message.data);
            var CurSet = Self.m_MapSet.get(CurMsg.Key);
            if (undefined === CurSet) {
                Self.m_MapSet.set(CurMsg.Key, {
                    Stats : 0,
                    Msgs : [
                        {
                            Message : message,
                            Headers : headers,
                            DeliveryInfo : deliveryInfo,
                            MessageObject : messageObject,
                            CreatedAt : (new Date()).getTime()
                        }
                    ]
                });
            } else {
                CurSet.Msgs.push({
                    Message : message,
                    Headers : headers,
                    DeliveryInfo : deliveryInfo,
                    MessageObject : messageObject,
                    CreatedAt : (new Date()).getTime()
                });
            }
            if (Self.m_CurConc < Self.m_MaxConc) {
                setTimeout(() => {
                    Self._Upgrade();
                }, 0);
            }
        });
    }

    Upgrade(complate, key) {
        if (complate === true) {
            this.m_CurConc--;
            var CurSet = this.m_MapSet.get(key);
            if (CurSet) {
                CurSet.Stats = 0;
            }
        }
        var Self = this;
        setTimeout(() => {
            Self._Upgrade();
        }, 0);
    }

    _Upgrade() {
        if (this.m_CurConc < this.m_MaxConc) {
            // console.log(this.m_CurConc);
            for (var item of this.m_MapSet.entries()) {
                if (item[1].Stats === 0 && item[1].Msgs.length > 0) {
                    var CurMsg = item[1].Msgs.shift();
                    if (TypeUtils.isFunction(this.m_CB)) {
                        this.m_CurConc++;
                        item[1].Stats = 1;
                        // console.log('xxx');
                        this.m_CB(item[0], CurMsg.Message, CurMsg.Headers, CurMsg.DeliveryInfo, CurMsg.MessageObject);
                        if (this.m_CurConc >= this.m_MaxConc) {
                            return;
                        }
                    }
                }
            }
            // for 清理
        }
    }

    Shift(reject, requeue) {
        if (TypeUtils.isEmptyObj(this.m_Queue) === true) {
            throw 'No init';
        }

        this.m_Queue.shift(reject, requeue);
    }

    Next() {
        this.m_Idle = true;
        var Self = this;
        setTimeout(() => {
            Self._Next();
        }, 0);
    }

    _Next() {
        if (this.m_TempQueue.length !== 0) {
            var CurMsg = this.m_TempQueue.shift();
            if (TypeUtils.isFunction(this.m_CB)) {
                this.m_Idle = false;
                // console.log('xxx');
                this.m_CB(CurMsg.Message, CurMsg.Headers, CurMsg.DeliveryInfo, CurMsg.MessageObject);
            }
        }
    }
}

module.exports  = MQSubscriber;
