
import fs = require('fs');
import crypto = require('crypto');

/**
 * 获取UUID十六进制字符串
 * @param keyBytes 长度,默认16字节
 * @returns 不带横线的16字节,32位,hex字符串
 */
export function getUUIDHexStr(keyBytes?: number) {
    return crypto.randomBytes(keyBytes ? keyBytes : 16).toString('hex');
}


/**
 * 获取UUID十六进制字符串
 * @param keyBytes 长度,默认16字节
 * @returns 不带横线的16字节,32位,hex字符串
 */
export function getUUID(keyBytes?: number) {
    return crypto.randomBytes(keyBytes ? keyBytes : 16).toString('hex');
}

/**
 * 获取md5
 * @param data 数据体
 * @returns 计算出的md5
 */
export function getMd5(data: string | Buffer) {
    var md5 = crypto.createHash("md5");
    md5.update(data);
    return md5.digest('hex');
}

/**
 * 字符串转证书文件
 * @param kayStr 密钥字符串
 * @param mode 密钥类型,1=pem公钥,2=pem私钥
 * @returns pem证书字符串
 */
export function keyStr2pemStr(kayStr: string, mode: number) {
    var bodyStr = "";
    for (var i = 0; i < 100; i++) {
        var subStr = kayStr.substr(i * 64, 64);
        if (!subStr) break;
        bodyStr += kayStr.substr(i * 64, 64) + "\n";
    }

    if (mode === 1) {
        return "-----BEGIN PUBLIC KEY-----\n" + bodyStr + "-----END PUBLIC KEY-----";
    }
    else if (mode === 2) {
        return "-----BEGIN RSA PRIVATE KEY-----\n" + bodyStr + "-----END RSA PRIVATE KEY-----";
    }
    return "err mode";
}

/**
 * 获取文件MD5
 * @param filePath 文件路径
 * @returns MD5结果,16字节hex小写字符串
 */
export function getFileMd5(filePath: string) {
    return new Promise<string>((resolve, reject) => { // Begin Promise

        var rs = fs.createReadStream(filePath);

        var hash = crypto.createHash('md5');
        rs.on('data', hash.update.bind(hash));

        rs.on('end', function () {
            resolve(hash.digest('hex'));
        });

        rs.on('error', function (err) {
            reject(err);
        });

    });
}

/**
 * 加密数据(Java对应的是:[AES/DES/desede]/[ECB/CBC]/PKCS5Padding)
 * @param algorithm 算法:
 * AES模式:aes-[128/192/256]-[ecb/cbc] 密钥16字节,iv16字节
 * DES模式:des-[ecb/cbc] 密钥8字节,iv8字节
 * 3DES模式:des-[ede3/ede3-cbc] 密钥24字节(16字节:前8+后8+前8=24字节),iv8字节
 * @param data 需要加密的数据
 * @param key 密钥,可以用hex字符串
 * @param iv 起始向量,用于CBC等混杂,无需混杂用0填充,可以用hex字符串,没有可以undefined
 * @param noPadding 不使用填(省略为pk5/7填充)
 */
export function encode(algorithm: string, data: Buffer, key?: string | Buffer, iv?: string | Buffer, noPadding?: boolean) {
    // 参数处理
    if (typeof (key) === "string") {
        if (<any>Buffer.alloc) {
            key = Buffer.from(key, "hex");
        }
        else {
            key = new Buffer(key, "hex");
        }
    }
    if (!iv) iv = "";
    if (typeof (iv) === "string") {
        if (<any>Buffer.alloc) {
            iv = Buffer.from(iv, "hex");
        }
        else {
            iv = new Buffer(iv, "hex");
        }
    }

    // 数据处理
    var coder = crypto.createCipheriv(algorithm, <any>key, <any>iv);
    if (noPadding) coder.setAutoPadding(false);
    var crypted = coder.update(data);
    return Buffer.concat([crypted, coder.final()]);
}

/**
 * 解密数据
 * @param algorithm 算法:
 * AES模式:aes-[128/192/256]-[ecb/cbc] 密钥16字节,iv16字节
 * DES模式:des-[ecb/cbc] 密钥8字节,iv8字节
 * 3DES模式:des-[ede3/ede3-cbc] 密钥24字节(16字节:前8+后8+前8=24字节),iv8字节
 * @param data 需要加密的数据
 * @param key 密钥可以用hex字符串
 * @param iv 起始向量,用于CBC等混杂,无需混杂用0填充,可以用hex字符串
 * @param noPadding 不使用填(省略为pk5/7填充)
 */
export function decode(algorithm: string, data: Buffer, key?: string | Buffer, iv?: string | Buffer, noPadding?: boolean) {
    // 参数处理
    if (typeof (key) === "string") {
        if (<any>Buffer.alloc) {
            key = Buffer.from(key, "hex");
        }
        else {
            key = new Buffer(key, "hex");
        }
    }
    if (!iv) iv = "";
    if (typeof (iv) === "string") {
        if (<any>Buffer.alloc) {
            iv = Buffer.from(iv, "hex");
        }
        else {
            iv = new Buffer(iv, "hex");
        }
    }

    // 数据处理
    var coder = crypto.createDecipheriv(algorithm, <any>key, <any>iv);
    if (noPadding) coder.setAutoPadding(false);
    var crypted = coder.update(data);
    return Buffer.concat([crypted, coder.final()]);
}

