
import fs = require('fs');
import path = require('path');
import os = require('os');
import jco = require("./jco");
import jbase_utils = require("./jbase_utils");

import Jlog = require('./jlog');
var jlog = Jlog("jfile_utils");


/**
 * 文件操作错误返回
 * @param error 错误对象
 */
export type TypeDoFileCbFun = (error: Error | null) => void;

/**
 * fs的copyFile实现
 * @param sourceFile 源文件
 * @param destFile 目标文件
 * @param ignoreExist 跳过存在的目标文件并抛错
 * @param cbFun 回调
 */
export function fsCopyFile(sourceFile: string, destFile: string, ignoreExist?: boolean, cbFun?: TypeDoFileCbFun) {

    // 是否存在系统函数
    if (fs.copyFile) {
        fs.copyFile(sourceFile, destFile, ignoreExist ? fs.constants.COPYFILE_EXCL : 0, <any>cbFun);
        return;
    }

    // 检查文件是否存在
    fs.lstat(destFile, (lsErr, res) => {
        if (!lsErr && ignoreExist) {
            // 文件已存且跳过
            if (cbFun) {
                var newErr = <any>new Error("EEXIST: file already exists, copyfile");
                newErr.errno = -4075,
                    newErr.code = 'EEXIST',
                    newErr.syscall = 'copyfile',
                    newErr.path = sourceFile,
                    newErr.dest = destFile
                cbFun(newErr);
            }
        }
        else {

            // 完成标志
            var jfinish = false;
            // 读流
            var readStream = fs.createReadStream(sourceFile);

            // 写流
            var tmpFile = destFile + "_" + Math.random();
            var writeStream = fs.createWriteStream(tmpFile);

            // 关联管道
            readStream.pipe(writeStream);

            // 捕获读流错误
            readStream.once("error", function (err) {
                if (jfinish) return;
                jfinish = true;
                fs.unlink(tmpFile, (err) => { });
                if (cbFun) cbFun(err);
            });

            // 捕获写流错误
            writeStream.once("error", function (err) {
                if (jfinish) return;
                jfinish = true;
                fs.unlink(tmpFile, (err) => { });
                if (cbFun) cbFun(err);
            });

            // 捕获写关闭
            writeStream.once("close", function () {
                if (jfinish) return;
                jfinish = true;
                if (!lsErr) {
                    // 删除源目标文件
                    fs.unlink(destFile, (unLinkErr) => {
                        if (unLinkErr) {
                            fs.unlink(tmpFile, (err) => { });
                            if (cbFun) cbFun(unLinkErr || null);
                        }
                        else {
                            // 移动
                            fs.rename(tmpFile, destFile, (errRnm) => {
                                if (cbFun) cbFun(errRnm || null);
                            })
                        }
                    })
                }
                else {
                    // 移动
                    fs.rename(tmpFile, destFile, (errRnm) => {
                        if (cbFun) cbFun(errRnm || null);
                    })
                }

            });

        }
    });
}

/**
 * fs的moveFile实现
 * @param sourceFile 源文件
 * @param destFile 目标文件
 * @param cbFun 回调
 */
export function fsMoveFile(sourceFile: string, destFile: string, cbFun?: TypeDoFileCbFun) {

    // 先快速移动
    fs.rename(sourceFile, destFile, function (err1) {

        // 不同磁盘使用管道复制再删除
        if (err1) {
            fsCopyFile(sourceFile, destFile, true, function (err2) {
                if (err2) {
                    // 失败
                    if (cbFun) cbFun(err2);
                }
                else {
                    // 删除旧文件
                    fs.unlink(sourceFile, function (err3) {
                        if (err3) {
                            // 删除失败
                            if (cbFun) cbFun(<any>err3);
                        }
                        else {
                            // 成功
                            if (cbFun) cbFun(null);
                        }
                    });
                }
            });
        }
        else {
            // 快速移动成功
            if (cbFun) cbFun(null);
        }

    });
}

/**
 * 创建文件夹
 * @param dirPath 路径
 * @param mode 模式
 */
export function mkdir(dirPath: string, mode?: any) {
    return new Promise<boolean>((resolve, reject) => { // Begin Promise
        fs.mkdir(dirPath, mode, (err) => {
            if (err && jlog.jlogEnable) jlog.trace("mkDir err:", err); // 错误
            return err ? resolve(false) : resolve(true);
        })
    }); // End Promise
}

/**
 * 复制文件(不抛异常,只返回true/false)
 * @param oldPath 源文件
 * @param newPath 目标文件
 * @param ignoreExist 跳过存在的目标文件并返回false
 */
export function copyFile(oldPath: string, newPath: string, ignoreExist?: boolean) {
    return new Promise<boolean>((resolve, reject) => { // Begin Promise
        fsCopyFile(oldPath, newPath, ignoreExist, (err) => {
            if (err && jlog.jlogEnable) jlog.trace("copyFile err:", err); // 错误
            return err ? resolve(false) : resolve(true);
        })
    }); // End Promise
}

/**
 * 移动文件(不抛异常,只返回true/false)
 * @param oldPath 源文件
 * @param newPath 目标文件
 */
export function moveFile(oldPath: string, newPath: string) {
    return new Promise<boolean>((resolve, reject) => { // Begin Promise
        fsMoveFile(oldPath, newPath, (err) => {
            if (err && jlog.jlogEnable) jlog.trace("MoveFile err:", err); // 错误
            return err ? resolve(false) : resolve(true);
        })
    }); // End Promise
}


/**
 * 删除文件(不抛异常,只返回true/false)
 * @param filePath 源文件
 */
export function deleteFile(filePath: string) {
    return new Promise<boolean>((resolve, reject) => { // Begin Promise
        fs.unlink(filePath, (err) => {
            if (err && jlog.jlogEnable) jlog.trace("deleteFile err:", err); // 错误
            return err ? resolve(false) : resolve(true);
        })
    }); // End Promise
}

/**
 * 删除文件
 * @param filePath 源文件
 */
export function unlink(filePath: string): Promise<any> {
    return new Promise<void>((resolve, reject) => {
        fs.unlink(filePath, (err) => {
            return err ? reject(err) : resolve();
        })
    });
}

/**
 * 创建多级目录
 * @param dirPath 目录
 * @param noLast 去掉最后一级
 * @param mode 模式,用于权限修改等
 * @returns 无
 */
export function createDirs(dirPath: string, noLast?: boolean, mode?: any) {
    var nPath = path.normalize(dirPath); // 格式化路径
    if (noLast) nPath = path.dirname(nPath); // 退一层

    var i;

    // 目录层
    var dirArr = [];
    var lastDir = nPath;
    for (i = 0; i < 100; i++) {
        nPath = path.dirname(nPath);
        if (lastDir === nPath) break;
        dirArr.push(lastDir);
        lastDir = nPath;
    }

    // 建立目录
    for (i = dirArr.length - 1; i >= 0; i--) {
        try {
            fs.mkdirSync(dirArr[i], mode);
        } catch (error) {
        }
    }
}

/**
 * 读取整个文件
 * @param path 路径
 * @param option 选项,可以是utf8等编码(可省)
 * @returns 文件内容,buffer类型,失败返回undefined
 */
export function readFile(path: string, option?: any) {
    return new Promise<Buffer | string>((resolve, reject) => { // Begin Promise
        fs.readFile(path, option, (err, data) => {
            if (err && jlog.jlogEnable) jlog.trace("readFile err:", err); // 错误
            return err ? reject(err) : resolve(data);
        });
    }); // End Promise
}


/**
 * 写出整个文件
 * @param path 路径
 * @param data 数据
 * @param option 选项,可以是utf8等编码(可省)
 * @returns 返回布尔值
 */
export function writeFile(path: string, data: any, option?: any) {
    return new Promise<void>((resolve, reject) => { // Begin Promise
        fs.writeFile(path, data, option, (err) => {
            if (err && jlog.jlogEnable) jlog.trace("writeFile err:", err); // 错误
            return err ? reject(err) : resolve();
        });
    }); // End Promise
}

/**
 * 打开文件
 * @param path 路径
 * @param flags 标记,如同C++无b,例:r/w/a+等
 * @param mode 文件权限,8进制,默认:0o666,可省
 * @returns 句柄
 */
export function open(path: fs.PathLike, flags: string | number, mode?: string | number) {
    return new Promise<number>((resolve, reject) => { // Begin Promise
        fs.open(path, flags, mode, (err, fd) => {
            if (err && jlog.jlogEnable) jlog.trace("open err:", err); // 错误
            return err ? reject(err) : resolve(fd);
        });
    }); // End Promise
}

/**
 * 文件读取
 * @param fd 句柄
 * @param buffer 接收缓存
 * @param offset 接收偏移
 * @param length 接收长度
 * @param position 文件指针
 * @returns 读取到的长度
 */
export function read(fd: number, buffer: Buffer, offset: number, length: number, position?: number) {
    return new Promise<number>((resolve, reject) => { // Begin Promise
        fs.read(fd, buffer, offset, length, (position === undefined) ? null : position, (err, bytesRead) => {
            if (err && jlog.jlogEnable) jlog.trace("read err:", err); // 错误
            return err ? reject(err) : resolve(bytesRead);
        });
    }); // End Promise
}

/**
 * 文件写入
 * @param fd 文件句柄
 * @param buf_str 缓存/字符串
 * @param bufOffset_strPosition 缓存的起始偏移/字符串的文件写入指针
 * @param bufLength_strEncoding 缓存长度/字符串类型编码
 * @param bufPosition 缓存的文件写入指针/字符串不填
 * @returns 写入的长度
 */
export function write(fd: number, buf_str: Buffer | string, bufOffset_strPosition?: number, bufLength_strEncoding?: number | string, bufPosition?: number) {
    return new Promise<number>((resolve, reject) => { // Begin Promise
        if (typeof (buf_str) === "string") {
            fs.write(fd, <string>buf_str, bufOffset_strPosition, <any>bufLength_strEncoding, (err, written) => {
                if (err && jlog.jlogEnable) jlog.trace("write err:", err); // 错误
                return err ? reject(err) : resolve(written);
            });
        }
        else {
            fs.write(fd, <Buffer>buf_str, bufOffset_strPosition, <number>bufLength_strEncoding, bufPosition, (err, written) => {
                if (err && jlog.jlogEnable) jlog.trace("write err:", err); // 错误
                return err ? reject(err) : resolve(written);
            });
        }
    }); // End Promise
}

/**
 * 关闭文件
 * @param fd 句柄
 * @returns 无
 */
export function close(fd: number) {
    return new Promise<void>((resolve, reject) => { // Begin Promise
        fs.close(fd, (err) => {
            if (err && jlog.jlogEnable) jlog.trace("close err:", err); // 错误
            return err ? reject(err) : resolve();
        });
    }); // End Promise
}



/**
 * 流文件打开写流
 * @param path 路径
 * @param options 选项
 */
export function openSw(path: fs.PathLike, options?: string | {
    flags?: string;
    encoding?: any;
    mode?: number;
    start?: number;
}) {
    return new Promise<fs.WriteStream>((resolve, reject) => { // Begin Promise

        // 创建流
        var stFd = fs.createWriteStream(path, <any>options);
        var stFdany = <any>stFd;

        // 关闭句柄
        stFdany.jclose = new Promise((resolve, reject) => { // Begin Promise
            stFdany.jclose_resolve = resolve;
            stFdany.jclose_reject = reject;
        }); // End Promise

        // 错误事件
        stFd.on("error", (err) => {
            reject(err);
            stFdany.jclose_reject(err);
        });

        // 打开事件
        stFd.on("open", (fd) => {
            resolve(stFd);
        });

        // 关闭事件
        stFd.on("close", () => {
            stFdany.jclose_resolve();
        });

    }); // End Promise
}

/**
 * 流文件写入(快速写入)
 * @param stFd 流对象
 * @param data 数据
 * @param encoding 字符串编码(仅字符串使用)
 */
export function writeSw(stFd: fs.WriteStream, data: string | Buffer, encoding?: any) {

    // 写入
    var ret: boolean;
    if (typeof (data) === "string") {
        ret = stFd.write(data, encoding || "utf8");
    }
    else {
        ret = stFd.write(data);
    }

    // 缓冲是否足够
    if (!ret) {
        return new Promise<void>((resolve, reject) => { // Begin Promise
            // 等待缓冲完成后
            stFd.once("drain", () => {
                resolve();
            });
        }); // End Promise
    }
    else {
        // 直接已决
        return Promise.resolve();
    }
}

/**
 * 关闭文件写流
 * @param stFd 流对象
 */
export function closeSw(stFd: fs.WriteStream) {
    stFd.end();
    return (<any>stFd).jclose;
}


/**
 * 获取临时目录
 * @param mode 模式,未定义或0=系统临时目录,1=当前路径的临时目录,2=代码目录的临时目录
 * @returns 临时目录的绝对路径,末尾带斜线
 */
export function getTempPath(mode: number) {
    var ret;
    if (mode === 1) {
        ret = process.cwd() + "/";
    } else if (mode === 2) {
        ret = __dirname + "/";
    }
    else {
        ret = os.tmpdir() + "/";
    }
    return path.normalize(ret);
}


export type readByLineCb = (strBuf: Buffer, strSize: number, lineNum: number) => void;

/**
 * 读取文件行
 * @param filePath 文件路径
 * @param cbFun 回调函数(strBuf,strSize,lineNum),可以使用Generator(速度慢)
 * @returns 行数
 */
export function readByLine(filePath: string, cbFun: readByLineCb) {

    return new Promise<number>((resolve, reject) => { // Begin Promise

        var lineCount = 0;
        var strBuff = <Buffer><any>null;
        if (<any>Buffer.alloc) {
            strBuff = Buffer.allocUnsafe(7168);
        }
        else {
            strBuff = new Buffer(7168);
        }
        var strLen = 0;

        // 异步支持
        var genMode = false;
        if (jbase_utils.isYieldFun(cbFun)) {
            genMode = true;
        }

        // 换行符
        var eol = os.EOL;
        if (eol.length > 1) eol = '\n';

        // 文件处理
        var rs = fs.createReadStream(filePath, { highWaterMark: 1024 });

        rs.on("data", function (chunk: Buffer) {
            var chkIndex = 0;
            var lineLast = -1;

            // 无回调函数
            if (!cbFun) return;

            // 是否异步
            if (genMode) {
                rs.pause();
                jco(function* (): any {

                    // 查找行尾
                    while (chkIndex < chunk.length) {
                        // 查找行尾
                        lineLast = chunk.indexOf(eol, chkIndex);
                        if (lineLast >= 0) {
                            // 复制到目标缓存
                            chunk.copy(strBuff, strLen, chkIndex, lineLast + 1);
                            strLen += lineLast + 1 - chkIndex;

                            // 操作
                            try {
                                if (lineCount === 0 && strLen >= 3 && strBuff[0] === 0xEF && strBuff[1] === 0xBB && strBuff[2] === 0xBF) {
                                    var buf_t = <Buffer><any>null;
                                    if (<any>Buffer.alloc) {
                                        buf_t = Buffer.allocUnsafe(strLen);
                                    }
                                    else {
                                        buf_t = new Buffer(strLen);
                                    }
                                    strBuff.copy(buf_t, 0, 3, strLen);
                                    yield cbFun(buf_t, strLen - 3, lineCount);
                                }
                                else {
                                    yield cbFun(strBuff, strLen, lineCount);
                                }
                            } catch (error) {
                                rs.destroy();
                                throw error;
                            }
                            lineCount++;

                            // 下一包
                            strLen = 0;
                            chkIndex = lineLast + 1;
                        }
                        else {
                            // 保留剩余
                            chunk.copy(strBuff, strLen, chkIndex);
                            strLen += chunk.length - chkIndex;
                            break;
                        }
                    }

                }).then(() => {
                    rs.resume();
                }).catch((error) => {
                    reject(error);
                })
            }
            else {
                // 查找行尾
                while (chkIndex < chunk.length) {
                    // 查找行尾
                    lineLast = chunk.indexOf(eol, chkIndex);
                    if (lineLast >= 0) {
                        // 复制到目标缓存
                        chunk.copy(strBuff, strLen, chkIndex, lineLast + 1);
                        strLen += lineLast + 1 - chkIndex;

                        // 操作
                        try {
                            if (lineCount === 0 && strLen >= 3 && strBuff[0] === 0xEF && strBuff[1] === 0xBB && strBuff[2] === 0xBF) {
                                var buf_t = <Buffer><any>null;
                                if (<any>Buffer.alloc) {
                                    buf_t = Buffer.allocUnsafe(strLen);
                                }
                                else {
                                    buf_t = new Buffer(strLen);
                                }
                                strBuff.copy(buf_t, 0, 3, strLen);
                                cbFun(buf_t, strLen - 3, lineCount);
                            }
                            else {
                                cbFun(strBuff, strLen, lineCount);
                            }
                        } catch (error) {
                            rs.destroy();
                            cbFun = <any>null; // 不再执行结束调用
                            reject(error);
                        }
                        lineCount++;

                        // 下一包
                        strLen = 0;
                        chkIndex = lineLast + 1;
                    }
                    else {
                        // 保留剩余
                        chunk.copy(strBuff, strLen, chkIndex);
                        strLen += chunk.length - chkIndex;
                        break;
                    }
                }
            }

        });

        // 结束
        rs.on("end", function () {
            // 最后一包
            if (genMode) {
                jco(function* (): any {
                    if (cbFun) {
                        yield cbFun(strBuff, strLen, lineCount)
                    }
                    lineCount++;
                }).then(() => {
                    resolve(lineCount);
                }).catch((err) => {
                    reject(err);
                });
            }
            else {
                if (cbFun) {
                    cbFun(strBuff, strLen, lineCount);
                }
                lineCount++;
                resolve(lineCount);
            }
        });

        // 错误
        rs.on("error", function (err) {
            reject(err);
        });

    }); // End Promise

}

/**
 * 读取目录
 * @param path 路径
 */
export function readdir(path: string) {
    return new Promise<Array<string>>((resolve, reject) => { // Begin Promise
        fs.readdir(path, (err, res) => {
            return err ? resolve([]) : resolve(res);
        });
    });
}


/**
 * 读取路径属性(符号连接当作符号文件)
 * @param path 路径
 */
export function lstat(path: string) {
    return new Promise<fs.Stats>((resolve, reject) => { // Begin Promise
        fs.lstat(path, (err, res) => {
            if (err && jlog.jlogEnable) jlog.trace("lstat err:", err); // 错误
            return err ? reject(err) : resolve(res);
        });
    });
}

/**
 * 读取路径属性
 * @param path 路径
 */
export function stat(path: string) {
    return new Promise<fs.Stats>((resolve, reject) => {
        fs.stat(path, (err, res) => {
            if (err && jlog.jlogEnable) jlog.trace("stat err:", err); // 错误
            return err ? reject(err) : resolve(res);
        });
    });
}


/**
 * 判断文件存在的属性(不抛异常,只返回true/false)
 * @param path 路径
 * @param mode 模式(省略为只检查是否存在),参考[fs.constants.W_OK可写]
 */
export function access(path: string, mode?: number) {
    return new Promise((resolve, reject) => { // Begin Promise
        fs.access(path, mode, (err) => {
            return err ? resolve(false) : resolve(true);
        });
    });
}

export type copyFolderRetType = {
    srcFiles: number,
    srcFolders: number,
    srcSize: number,
    destFiles: number,
    destFolders: number,
    destSize: number,
}

/**
 * 复制文件夹
 * @param srcDir 原始目录
 * @param destDir 目标目录
 * @param ignoreExist 是否忽略已存在
 * @param mode 目录权限模式
 * @param initInfo 附加信息(不要填写)
 */
function copyFolderEx(srcDir: string, destDir: string, ignoreExist?: boolean, mode?: any, initInfo?: any) {
    return <Promise<copyFolderRetType>>jco(function* (): any {
        if (!initInfo) initInfo = {
            srcFiles: 0,
            srcFolders: 0,
            srcSize: 0,
            destFiles: 0,
            destFolders: 0,
            destSize: 0
        };
        initInfo.srcFolders++;
        if ((yield mkdir(destDir, mode)) || !ignoreExist) {
            initInfo.destFolders++;
        }

        var fileList: Array<string> = yield readdir(srcDir);
        for (var index = 0; index < fileList.length; index++) {
            var fileName = fileList[index];
            var srcPath = srcDir + "/" + fileName;
            var destPath = destDir + "/" + fileName;
            var fileStat: fs.Stats;
            try {
                fileStat = yield lstat(srcPath);

                if (fileStat.isDirectory()) {
                    // 目录
                    yield copyFolderEx(srcPath, destPath, ignoreExist, mode, initInfo);
                }
                else {
                    // 文件
                    initInfo.srcFiles++;
                    initInfo.srcSize += fileStat.size;
                    if (yield copyFile(srcPath, destPath, ignoreExist)) {
                        initInfo.destFiles++;
                        initInfo.destSize += fileStat.size;
                    }
                }
            } catch (error) {
            }
        }

        return initInfo;
    });
}

/**
 * 复制文件夹
 * @param srcDir 原始目录
 * @param destDir 目标目录
 * @param ignoreExist 是否忽略已存在
 * @param mode 目录权限模式
 */
export function copyFolder(srcDir: string, destDir: string, ignoreExist?: boolean, mode?: any) {
    return copyFolderEx(srcDir, destDir, ignoreExist, mode);
}

/**
 * 删除空目录
 * @param srcDir 目录
 */
export function rmdir(srcDir: string): Promise<any> {
    return new Promise<void>((resolve, reject) => {
        fs.rmdir(srcDir, (err) => {
            return err ? reject(err) : resolve();
        })
    })
}


/**
 * 删除文件夹
 * @param srcDir 目录
 */
function deleteFolderEx(srcDir: string) {
    return <Promise<undefined>>jco(function* (): any {
        var fileList: Array<string> = yield readdir(srcDir);
        for (var index = 0; index < fileList.length; index++) {
            var fileName = fileList[index];
            var srcPath = srcDir + "/" + fileName;
            var fileStat: fs.Stats;
            try {
                fileStat = yield lstat(srcPath);
                if (fileStat.isDirectory()) {
                    // 目录
                    yield deleteFolderEx(srcPath);
                }
                else {
                    // 文件
                    yield deleteFile(srcPath);
                }
            } catch (error) {
            }
        }
        try {
            // 删除空目录
            yield rmdir(srcDir);
        } catch (error) {
        }
    });
}


/**
 * 删除文件夹
 * @param srcDir 目录
 */
export function deleteFolder(srcDir: string) {
    return deleteFolderEx(srcDir);
}


/**
 * 剪切路径
 * @param fileName 文件名
 * @param layers 忽略层数
 * @param noExName 去掉扩展名
 * @param nolast 去掉最后的文件名
 */
export function cutPath(fileName: string, layers: number, noExName?: boolean, nolast?: boolean) {
    fileName = path.resolve(fileName);
    if (nolast) fileName = path.dirname(fileName);
    var cwd = path.resolve(process.cwd());
    var str_t = fileName.substr(cwd.length, fileName.length - cwd.length - (noExName ? path.extname(fileName).length : 0)).replace(/\\/g, '/');
    var p1 = str_t.indexOf("/", 0);
    for (; p1 >= 0 && layers > 0; layers--) {
        p1 = str_t.indexOf("/", p1 + 1);
    }
    if (p1 < 0) {
        return ""
    }
    else {
        return "/" + str_t.substr(p1 + 1);
    }
}

