import fs = require("fs");
import jco = require("./jco");
import path = require("path");

import Jlog = require('./jlog');
var jlog = Jlog("jmod_loader");

/**
 * 载入函数回调定义
 */
export type LoadCbFun = (err: Error | null, res: { filePath: string, fileMainName: string, obj?: any }) => any;

/**
 * 列目录
 * @param dir 目录
 * @param cbFun 回调函数
 * @returns 无
 */
function* listDir(dir: string, cbFun: any): any {
    var fileList = fs.readdirSync(dir, 'utf-8');
    for (var i = 0; i < fileList.length; i++) {
        var stat = fs.lstatSync(dir + fileList[i]);
        if (stat.isDirectory()) {
            yield listDir(dir + fileList[i] + '/', cbFun); // 递归
        } else {
            yield cbFun(dir, fileList[i]); // 回调
        }
    }
}

/**
 * 加载初始化并返回对象
 * @param filePath 文件
 * @param fileMainName 文件主名
 * @param cbFun 错误回调
 * @returns 载入后的对象
 */
function* jrequireLoadInit(filePath: string, fileMainName: string, cbFun?: LoadCbFun): any {
    var loadObj = null;
    filePath = path.resolve(filePath);
    try {
        loadObj = require(filePath);
    } catch (error: any) {
        if (cbFun) {
            yield cbFun(error, { filePath: filePath, fileMainName: fileMainName });
        }
        else {
            if (jlog.jlogEnable) jlog.error("js require err:" + filePath, error);
        }
        return loadObj;
    }
    if (cbFun) {
        var newObj = yield cbFun(null, { filePath: filePath, fileMainName: fileMainName, obj: loadObj });
        if (newObj) loadObj = newObj; // 新的对象
    } else if (loadObj.init) {
        yield loadObj.init();
    }

    return loadObj;
}

/**
 * 载入所有模块并返回
 * @param dir 路径
 * @param useStrObj 使用字符串对象
 * @param cbFun 回调(error,成功后为载入的对象,存在回调则不会自动执行init方法)
 * @returns 无
 */
export function load(dir: string, useStrObj: boolean, cbFun?: LoadCbFun) {
    return jco(function* (): any {
        var retObj: any = {};
        dir = path.resolve(dir);
        dir += "/";
        yield listDir(dir, function* (curDir: string, curFilename: string): any {
            var loadObj;
            var fileMainName = curFilename.substr(0, curFilename.length - 3);
            var modePathStr = curDir.substr(dir.length);
            if (useStrObj) {
                if (curFilename.substr(curFilename.length - 3).toLowerCase() === ".js") {
                    loadObj = yield jrequireLoadInit(curDir + curFilename, fileMainName, cbFun);
                    var objName = modePathStr + fileMainName;
                    retObj[objName] = loadObj;
                }
            }
            else {
                var modPath = modePathStr.split("/");
                var curObj = retObj;
                for (var i = 0; i < modPath.length - 1; i++) {
                    if (modPath.length > 0) {
                        if (!curObj[modPath[i]]) curObj[modPath[i]] = {};
                        curObj = curObj[modPath[i]];
                    }
                }
                if (curFilename.substr(curFilename.length - 3) === ".js") {
                    loadObj = yield jrequireLoadInit(curDir + curFilename, fileMainName, cbFun);
                    curObj[curFilename.substr(0, curFilename.length - 3)] = loadObj;
                }
            }
        });
        return retObj;
    });
}

/**
 * 载入所有模块并返回,已废弃,推荐使用load函数
 * @param dir 路径
 * @param useStrObj 使用字符串对象
 * @param cbFun 回调(error,成功后为载入的对象,存在回调则不会自动执行init方法)
 * @returns 无
 */
export function loadAllModules(dir: string, useStrObj: boolean, cbFun?: LoadCbFun) {
    return load(dir, useStrObj, cbFun);
}


