"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getRetObj = getRetObj;
exports.chkRetObj = chkRetObj;
exports.setCode = setCode;
exports.addPageInfo = addPageInfo;
exports.getPageRet = getPageRet;
exports.getRetData = getRetData;
exports.getRetList = getRetList;
exports.getStdBody = getStdBody;
const jtype_utils = require("./jtype_utils");
const JERR_S = require("./jerror_type");
/**
 * 获取或修正返回参
 * @param obj // 对象,可以是int或者是{code:xxx,...}
 * @param msg // 提示文本
 * @param dataObj // 数据结构,如果有则obj参数应该为整数
 * @returns 返回对象
 */
function getRetObj(obj, msg, dataObj) {
    var retObj = {};
    // 判断code
    if (jtype_utils.getInt(obj) !== undefined) { // 是数值
        retObj = { code: obj };
        if (dataObj)
            retObj.data = dataObj;
    }
    else if (typeof (obj) === "string") { // 未定义错误,只使用消息
        retObj = { code: -1, msg: obj };
    }
    else if (typeof (obj) === "boolean") { // 布尔类型
        retObj = { code: obj ? 0 : -1 };
    }
    else if (typeof (obj) === "object") { // 是对象
        if (jtype_utils.getInt(obj.code) === undefined) {
            // 没有code码,是普通对象
            retObj = { code: 0, msg: "OK", data: obj };
        }
        else {
            // 自定义对象
            retObj = obj;
        }
    }
    else {
        // 无法返回类型
        retObj = { code: -1, msg: "Err ret type:" + Object.prototype.toString.call(obj), data: {} };
    }
    // 补充msg
    if (typeof (retObj.msg) !== "string") {
        retObj.msg = msg ? msg : JERR_S.getName(retObj.code);
    }
    // 补充对象
    if (typeof (retObj.data) !== "object") {
        retObj.data = {};
    }
    // 数组处理
    else if (jtype_utils.getArray(retObj.data)) {
        var dataList = retObj.data;
        retObj.data = {
            list: dataList
        };
    }
    // 添加页码信息
    if (retObj.data.list && jtype_utils.getArray(retObj.data.list) && !retObj.data.pagination) {
        addPageInfo(retObj.data, retObj.data.list.length, 1, retObj.data.list.length ? retObj.data.list.length : 10);
    }
    // 返回
    return retObj;
}
/**
 * 检测返回对象的格式,并修正缺失的msg和data
 * @param obj // 对象,必须是{code:xxx,...}
 * @param throwError // 格式异常抛错误
 * @returns 是否正确
 */
function chkRetObj(obj, throwError) {
    if ((typeof (obj) !== "object") || jtype_utils.getInt(obj.code) === undefined) { // 不是对象,或者对象没有code
        if (throwError) {
            throw JERR_S.newError(JERR_S.RET_FORMAT_ERR);
        }
        else {
            return false;
        }
    }
    // 修正剩余项
    getRetObj(obj);
    return true;
}
function setCode(obj, code) {
    obj.code = code;
}
/**
 * 添加分页信息
 * @param obj 处理的结构体,结构为{list:[...]}
 * @param allCount 总数
 * @param pageNum 页码(1~N)
 * @param pageSize 每页条数
 * @param curCount 当前条数(省略后将自动计算)
 * @returns 未定义
 */
function addPageInfo(obj, allCount, pageNum, pageSize, curCount) {
    if (!obj.pagination)
        obj.pagination = {};
    obj.pagination = { total: allCount, pageSize: pageSize, current: pageNum };
}
/**
 * 获取分页
 * @param objArry 数组对象
 * @param pageInfo 页信息{pageNum:页码1~N,pageSize:页大小}
 * @param rowFun 行处理函数(可省,行处理);(row,line)=>{};返回对象表示有效的对象(返回null且funDoAll为true表示跳过,否则为源对象)
 * @param funDoAll 是否所有行都进行处理(可省,若存在则会遍历每一条行信息,根据返回对象进行重组分页)
 * @returns 返回data对象体,包括list及分页信息
 */
function getPageRet(objArry, pageInfo, rowFun, funDoAll) {
    if (typeof (pageInfo) !== "object") {
        pageInfo = {};
    }
    var pageNum = pageInfo.pageNum ? pageInfo.pageNum : (pageInfo.current ? pageInfo.current : 1);
    var pageSize = pageInfo.pageSize ? pageInfo.pageSize : 30;
    var i;
    var retObj = { list: [] };
    var startIndex = (pageNum - 1) * pageSize;
    var listIndex = 0;
    var allCount = 0;
    var curObj;
    if (rowFun) { // 函数式
        if (funDoAll) { // 所有都处理
            for (i = 0; i < objArry.length; i++) {
                // 函数处理
                curObj = rowFun(objArry[i], i);
                if (curObj) {
                    // 有效
                    if (allCount >= startIndex && listIndex < pageSize) {
                        retObj.list[listIndex++] = curObj;
                    }
                    // 索引增加
                    allCount++;
                }
            }
        }
        else { // 按照原表处理
            allCount = objArry.length;
            for (i = startIndex; i < (startIndex + pageSize); i++) {
                if (i >= allCount)
                    break;
                curObj = rowFun(objArry[i], i);
                retObj.list[listIndex++] = curObj ? curObj : objArry[i];
            }
        }
    }
    else { // 默认数组
        allCount = objArry.length;
        for (i = startIndex; i < (startIndex + pageSize); i++) {
            if (i >= allCount)
                break;
            retObj.list[listIndex++] = objArry[i];
        }
    }
    // 添加页码信息
    addPageInfo(retObj, allCount, pageNum, pageSize);
    // 返回
    return retObj;
}
/**
 * 获取数据体
 * @param obj // 对象,必须是{code:xxx,...}
 * @returns 成功后返回数据体,失败返回null
 */
function getRetData(obj) {
    if (typeof (obj) === "object" && obj.code === 0 && typeof (obj.data) === "object") {
        return obj.data;
    }
    else {
        return null;
    }
}
/**
 * 获取数据体的List对象
 * @param obj // 对象,必须是{code:xxx,...}
 * @returns 成功后返回数据体,失败返回null
 */
function getRetList(obj) {
    if (typeof (obj) === "object" && obj.code === 0 && typeof (obj.data) === "object" && jtype_utils.getArray(obj.data.list)) {
        return obj.data.list;
    }
    else {
        return null;
    }
}
/**
 * 获取标准结构体对象
 * @param obj // 对象,是{code:xxx,...}
 * @param errRet // 错误时返回的对象,若不填则抛出Error
 * @returns 成功后返回数据体,失败返回null
 */
function getStdBody(obj, errRet) {
    if (obj && (obj.code == 0 || obj.code == 600 || obj.code == 200)) {
        // 抛出内部对象
        return obj.data ? obj.data : {};
    }
    else if (errRet !== undefined) {
        return errRet;
    }
    // 抛错
    if (!obj)
        obj = {};
    var err = new Error("[" + obj.code + "]" + ((typeof (obj.msg) === 'string') ? obj.msg : ""));
    err.errno = obj.code;
    err.body = obj;
    throw err;
}
