
import jco = require('./jco');
import jbase_utils = require("./jbase_utils");
import jtype_utils = require("./jtype_utils");

import Jlog = require('./jlog');
var jlog = Jlog("jsig_utils");


// 当前信号ID
var envIdCur = 0;

// 信号ID表
var sigWaitObj: any = {};

// 信号返回表
var sigRetObj: any = {};

// 事件对象
var envListObj = <{ [key: string]: Function[] }>{};

/**
* 初始化
* @returns 无
*/
function init() {
    // 初始化ID
    envIdCur = Math.floor(Math.random() * 100000) + 100000;
}

/**
 * 信号回调
 * @param envId 信号ID
 * @param ackObj 应答对象(不允许为undefined否则会与wait返回冲突,但可以是null)
 * @returns 无
 */
export function sigAck(envId: string | number, ackObj: any) {
    // 初始化
    if (envIdCur === 0) {
        init();
    }

    // 存在任务
    if (sigWaitObj[envId]) {
        if (sigWaitObj[envId].timeOut > 0) {
            // 正常未超时
            sigWaitObj[envId].timeOut = 0; // 取消超时
            sigWaitObj[envId].resolve(ackObj); // 成功
        }
        else {
            // 超时未设置,提前完成
            sigRetObj[envId] = ackObj;
        }
    }
}

/**
 * 信号等待函数声明
 */
type SigWaitCbFun = (envId: string | number) => void;

/**
 * 信号等待
 * @param timeOutMs 超时时间(ms),若小于100ms则视为秒,-1=永久等待,不能为0
 * @param envId 信号ID,可以是字符串,省略后将自动生成一个整数型envId
 * @param firstFun 提前处理函数,将传入envId,函数:(envId)=>{}
 * @returns 返回的数据(超时后将返回undefined)
 */
export function sigWait(timeOutMs: number, envId: string | number | null, firstFun?: SigWaitCbFun) {
    // 初始化
    if (envIdCur === 0) {
        init();
    }

    // 时间计算
    if (!timeOutMs) {
        timeOutMs = 1000;
    }
    else if (timeOutMs < 0) {
        timeOutMs = 0;
    }
    else if (timeOutMs < 100) {
        timeOutMs = timeOutMs * 1000; // 秒转换为ms
    }

    // 取出标记ID并执行动作
    var sigId: string | number = ++envIdCur;
    if (envId) sigId = envId;
    var sid = envIdCur;

    // 超时句柄
    var lastTimer = <any>null;

    // 委托处理
    return new Promise<any>((resolve, reject) => { // Begin Promise
        // 入列标记,如果存在旧的则冲掉
        var lastWObj = sigWaitObj[sigId];
        if (lastWObj && lastWObj.resolve) {
            lastWObj.resolve(undefined);
        }

        // 删除旧值
        delete sigRetObj[sigId];

        // 标记
        sigWaitObj[sigId] = { timeOut: 0, sid: sid };

        // 执行预处理
        if (firstFun && jbase_utils.isYieldFun(firstFun)) {
            // 异步函数
            jco(firstFun(sigId)).then(() => {
                // 判断是否已经完成
                if (sigRetObj[sigId] !== undefined) {
                    return resolve(sigRetObj[sigId]);
                }

                // 入列标记,如果存在旧的则冲掉,这里是为了防止异步时进来新的
                var lastWObj = sigWaitObj[sigId];
                if (lastWObj && lastWObj.resolve) {
                    lastWObj.resolve(undefined);
                }

                sigWaitObj[sigId] = { timeOut: 0, resolve: resolve, sid: sid };

                // 定时器
                if (timeOutMs) {
                    sigWaitObj[sigId].timeOut = timeOutMs;
                    lastTimer = setTimeout(() => {
                        lastTimer = null;
                        resolve(undefined); // 超时,返回undefined
                    }, timeOutMs);
                }
                else {
                    // 无限等待必须有timeOut
                    sigWaitObj[sigId].timeOut = 1;
                }

            }).catch((err) => {
                reject(err);
            });
        }
        else {
            // 标准函数
            if (firstFun) firstFun(sigId);

            // 判断是否已经完成
            if (sigRetObj[sigId] !== undefined) {
                return resolve(sigRetObj[sigId]);
            }

            sigWaitObj[sigId] = { timeOut: 0, resolve: resolve, sid: sid };

            // 定时器
            if (timeOutMs) {
                sigWaitObj[sigId].timeOut = timeOutMs;
                lastTimer = setTimeout(() => {
                    lastTimer = null;
                    resolve(undefined); // 超时,返回undefined
                }, timeOutMs);
            }
            else {
                // 无限等待必须有timeOut
                sigWaitObj[sigId].timeOut = 1;
            }
        }

    }).then((res) => {
        // 清理定时器
        if (lastTimer) {
            clearTimeout(lastTimer);
            lastTimer = null;
        }

        // 清理等待对象,只能清理自己的
        if (sigWaitObj[sigId] && sigWaitObj[sigId].sid === sid) {
            delete sigWaitObj[sigId];
            // 通常多重sid进入后这种事情必然不是唯一的,所以在这里删没毛病
            delete sigRetObj[sigId];
        }
        return res;
    });
}


/**
 * 事件等待(默认是有信号,不等待)
 * @param envName 事件名
 * @param timeOutMs 最长等待时间(ms),-1=无限等待,0=信号探测
 * @returns true=有信号,false=无信号(仅超时或者信号探测时返回false)
 */
export function envWait(envName: string, timeOutMs: number) {

    // 初始化
    if (envIdCur === 0) {
        init();
    }

    // 有信号
    if (!envListObj[envName]) {
        return Promise.resolve(true);
    }

    // 0=检测式,直接返回无信号
    if (!timeOutMs) {
        return Promise.resolve(false);
    }

    // 超时句柄
    var lastTimer = <any>null;

    // 无信号,等
    return new Promise((resolve, reject) => { // Begin Promise
        envListObj[envName].push(resolve);

        // 延时等待
        if (timeOutMs > 0) {
            lastTimer = setTimeout(() => {
                lastTimer = null;
                resolve(false); // 超时
            }, timeOutMs);
        }
    }).then((ret) => {
        if (lastTimer) {
            clearTimeout(lastTimer);
            lastTimer = null;
        }
        return ret;
    }); // End Promise

}


/**
 * 设置事件状态,如果不存在则新建(非Promise)
 * @param envName 事件名
 * @param isEnv 是否激活事件,激活后所有等待的时间将继续执行
 * @returns 等待函数个数
 */
export function envSet(envName: string, isEnv: boolean) {
    // 初始化
    if (envIdCur === 0) {
        init();
    }

    var retCount = 0;
    if (isEnv) {
        // 设置信号,释放Promise,删除数组
        if (envListObj[envName]) {
            retCount = envListObj[envName].length;
            for (var index = 0; index < retCount; index++) {
                envListObj[envName][index](true);
            }

            // 不再进入列表
            delete envListObj[envName];
        }
    }
    else {
        //取消信号,进入pending
        if (!envListObj[envName]) {
            envListObj[envName] = [];
        }
        else {
            retCount = envListObj[envName].length;
        }
    }

    return retCount;
}




export type ParamListDoCb = (err: Error | null, res: any, argInfo: ParamListDoTask, taskId: number) => void;
export type ParamListDoTask = {

    /**
     * 名称
     */
    name?: string,

    /**
     * 执行函数
     */
    fun: Function,

    /**
     * 执行函数参数数组
     */
    args: Array<any>,

    /**
     * this对象
     */
    thisObj?: any,

    /**
     * 执行回调
     */
    cbFun?: ParamListDoCb

    /**
     * 用户对象
     */
    userObj?: any,

    /**
     * 分组ID,设置后相同ID将只会在同一个队列执行
     */
    groupId?: number

};


/**
 * 任务列表对象
 */
var taskListObj = <{ [key: string]: Array<ParamListDoTask> }>{}


/**
 * 任务列表指令,0=正常,1=任务结束后停止,2=强行停止
 */
var taskListCmd = <{ [key: string]: number }>{}


/**
 * 自动超时对象
 */
var autoTimerObj = <{ [key: string]: { timer?: NodeJS.Timeout, cbFun?: Function } }>{};

/**
 * 启动多任务列表并发处理,将会创建后台的链表任务处理
 * @param taskName 任务名称
 * @param startCount 并发数
 * @param doCb 操作回调
 * @returns 总任务结束
 */
export function taskListStart(taskName: string, startCount: number, doCb?: ParamListDoCb): Promise<any> {

    // 判断队列
    if (taskListObj[taskName]) {
        return Promise.reject(new Error("taskListDo task exist:" + taskName));
    }

    // 建立队列
    taskListObj[taskName] = [];
    taskListCmd[taskName] = 0;

    // 等待队列
    var taskList = taskListObj[taskName];

    // 建立等待事件
    var ernName = "taskListDo_" + taskName;
    envSet(ernName, false);

    // 并发任务
    var retTaskList = [];
    for (var index = 0; index < startCount; index++) {

        // 建立任务
        retTaskList.push(jco(function* (): any {

            // 当前任务
            var curTask = null;
            var taskId = index;
            var taskNameCur = taskName + "_" + taskId;
            if (jlog.jlogEnable) jlog.info("taskListDo[" + taskNameCur + "] start");

            // 循环
            while (true) {

                // 如果停止则退出
                if (taskListCmd[taskName]) return taskNameCur;

                // 等待
                yield envWait(ernName, -1);

                // 设置事件阻塞
                envSet(ernName, false);

                // 消耗掉所有任务
                while (true) {
                    // 清空任务
                    curTask = null;

                    // 如果强制停止则退出
                    if (taskListCmd[taskName] === 2) return taskNameCur;

                    // 取出一个
                    if (taskList[0]) {
                        var gId = jtype_utils.getNumber(taskList[0].groupId, -1);
                        if (gId >= 0 && (gId % startCount !== taskId)) break;
                        curTask = taskList.shift();
                    }
                    if (!curTask || !curTask.fun) break;

                    try {
                        // 执行
                        var retRes = yield curTask.fun.apply(curTask.thisObj ? curTask.thisObj : null, curTask.args);

                        //成功任务回调
                        if (curTask.cbFun) {
                            try {
                                if (jbase_utils.isYieldFun(curTask.cbFun)) {
                                    yield curTask.cbFun(null, retRes, curTask, taskId);
                                }
                                else {
                                    curTask.cbFun(null, retRes, curTask, taskId);
                                }
                            } catch (error) {
                                if (jlog.jlogEnable) jlog.error("taskListDo[" + taskNameCur + "] ", error);
                            }
                        }

                        //成功总回调
                        if (doCb) {
                            try {
                                if (jbase_utils.isYieldFun(doCb)) {
                                    yield doCb(null, retRes, curTask, taskId);
                                }
                                else {
                                    doCb(null, retRes, curTask, taskId);
                                }
                            } catch (error) {
                                if (jlog.jlogEnable) jlog.error("taskListDo[" + taskNameCur + "] ", error);
                            }
                        }

                    } catch (error: any) {
                        // 错误
                        if (jlog.jlogEnable) jlog.error("taskListDo[" + taskNameCur + "] ", error);

                        // 错误任务回调
                        if (curTask.cbFun) {
                            try {
                                if (jbase_utils.isYieldFun(curTask.cbFun)) {
                                    yield curTask.cbFun(error, null, curTask, taskId);
                                }
                                else {
                                    curTask.cbFun(error, null, curTask, taskId);
                                }
                            } catch (error) {
                                if (jlog.jlogEnable) jlog.error("taskListDo[" + taskNameCur + "] ", error);
                            }
                        }

                        // 错误总回调
                        if (doCb) {
                            try {
                                if (jbase_utils.isYieldFun(doCb)) {
                                    yield doCb(error, null, curTask, taskId);
                                }
                                else {
                                    doCb(error, null, curTask, taskId);
                                }
                            } catch (error) {
                                if (jlog.jlogEnable) jlog.error("taskListDo[" + taskNameCur + "] ", error);
                            }
                        }

                    }
                }

            }

        }).then((taskNameCur) => {
            if (jlog.jlogEnable) jlog.info("taskListDo[" + taskNameCur + "] stop");
        }))
    }

    return Promise.all(retTaskList).then(() => {
        // 删除队列
        delete taskListObj[taskName];
        // 删除指令
        delete taskListCmd[taskName];
    });
}

/**
 * 添加多任务列表并发任务
 * @param taskName 任务名称
 * @param taskObj 任务对象或任务数组
 */
export function taskListAdd(taskName: string, taskObj: jbase_utils.ParamMultiDoTask | jbase_utils.ParamMultiDoTask[]) {

    // 判断队列
    if (!taskListObj[taskName]) {
        jlog.errorThrow("taskListAdd no task run:" + taskName);
        return
    }
    if (jtype_utils.getArray(taskObj) === undefined) {
        // 添加单条
        taskListObj[taskName].push(<any>taskObj);
    }
    else {
        // 添加多条
        for (var index = 0; index < (<any>taskObj).length; index++) {
            taskListObj[taskName].push((<any>taskObj)[index]);
        }
    }

    // 启动事件
    envSet("taskListDo_" + taskName, true);
}

/**
 * 删除多任务列表并发任务
 * @param taskName 任务名称
 * @param subName 子任务名称,为空则删除所有未执行的任务
 */
export function taskListDel(taskName: string, subName?: string) {
    // 判断队列
    var curTask = taskListObj[taskName];
    if (!curTask) {
        return;
    }

    // 删除队列内容
    for (var index = 0; index < curTask.length; index++) {
        if (!subName || curTask[index].name === subName) {
            (<any>curTask[index].fun) = undefined;
        }
    }
}



/**
 * 停止任务,如果要等待任务,需要保存taskListStart的返回值
 * @param taskName 任务名称
 * @param force 强行停止(不等待任务结束)
 */
export function taskListStop(taskName: string, force?: boolean) {
    if (jlog.jlogEnable) jlog.info("taskListDo[" + taskName + "] stop cmd");

    // 停止命令
    if (taskListCmd[taskName] !== undefined) {
        taskListCmd[taskName] = force ? 2 : 1;
    }

    // 启动事件
    envSet("taskListDo_" + taskName, true);
}

/**
 * 自动超时复位,每次调用将重置超时时间,超时后调用函数
 * @param sName 会话名称
 * @param timeout 超时时间
 * @param timeoutCbFun 超时后的回调函数
 */
export function autoTimeoutRst(sName: string, timeout: number, timeoutCbFun: Function): void {
    var curObj = autoTimerObj[sName];
    if (!curObj) {
        curObj = {};
        autoTimerObj[sName] = curObj;
    }
    curObj.cbFun = timeoutCbFun;

    if (curObj.timer) {
        clearTimeout(curObj.timer);
        delete curObj.timer;
    }
    // 永久
    if (timeout < 0) {
        return;
    }
    // 设置定时器
    curObj.timer = setTimeout(function () {
        autoTimeoutClear(sName);
    }, timeout);
}

/**
 * 自动超时清除
 * @param sName 会话名称
 * @param notRunCb 不执行超时回调
 */
export function autoTimeoutClear(sName: string, notRunCb?: boolean): void {
    var curObj = autoTimerObj[sName];
    if (!curObj) return;
    if (curObj.timer) {
        clearTimeout(curObj.timer);
        delete curObj.timer;
    }
    var cbFun = curObj.cbFun;
    delete autoTimerObj[sName];
    if (cbFun && !notRunCb) {
        cbFun();
    }
}


